#!/usr/bin/env python3

from pdspy.dust import *
import numpy

iron = Dust()
iron.set_optical_constants_from_henn("optical_constants/iron.txt")
iron.set_density(7.87)

olivine = Dust()
olivine.set_optical_constants_from_henn("optical_constants/olivine.txt")
olivine.set_density(3.49)

orthopyroxene = Dust()
orthopyroxene.set_optical_constants_from_henn("optical_constants/orthopyroxene.txt")
orthopyroxene.set_density(3.4)

troilite = Dust()
troilite.set_optical_constants_from_henn("optical_constants/troilite.txt")
troilite.set_density(4.83)

organics = Dust()
organics.set_optical_constants_from_henn("optical_constants/organics.txt")
organics.set_density(1.5)

water_ice = Dust()
water_ice.set_optical_constants_from_henn("optical_constants/water_ice.txt")
water_ice.set_density(0.92)

silicates = Dust()
silicates.set_optical_constants_from_draine("optical_constants/astronomical_silicates.txt")
silicates.set_density(3.3)
silicates.calculate_optical_constants_on_wavelength_grid(water_ice.lam)

organics = Dust()
organics.set_optical_constants_from_henn("optical_constants/organics.txt")
organics.set_density(1.5)

species = [silicates,troilite,organics,water_ice]
#mass_fraction = numpy.array([6.4e-3,7.68e-4,2.13e-3,1.4e-3])
mass_fraction = numpy.array([3.41e-3,7.68e-4,4.13e-3,5.55e-3])
rho = numpy.array([silicates.rho,troilite.rho,organics.rho,water_ice.rho])
abundances = (mass_fraction/rho)/(mass_fraction/rho).sum()

dust = mix_dust(species, abundances)

amin = 0.005e-4
amax = 1.000e-1
pl = 3.5

dust.calculate_size_distribution_opacity(amin, amax, pl, \
        coat_volume_fraction=0.0)

dust.write('pollack_1mm.hdf5')



from ..constants.physics import c
from ..constants.astronomy import arcsec
from .libimaging import Image
from astropy.utils.exceptions import AstropyWarning
import astropy.io.fits as fits
import astropy.wcs as wcs
import warnings
import numpy

def readimfits(filename):
    
    # Open the fits file.

    data = fits.open(filename)

    # Figure out the dimensions of each axis and create an array to put the data
    # into, and put the data into that array.

    try:
        npol, nfreq, ny, nx = data[0].data.shape
    except:
        nfreq, ny, nx = data[0].data.shape
        npol = 1
    
    image = numpy.zeros((ny,nx,nfreq,npol))

    for i in range(npol):
        for j in range(nfreq):
            try:
                image[:,:,j,i] = data[0].data[i,j,:,:].reshape(ny,nx)
            except:
                image[:,:,j,i] = data[0].data[j,:,:].reshape(ny,nx)

    # Read in the x and y coordinate information, including the WCS info if it
    # is available.

    header = data[0].header

    # Check whether there is a CASA beam table.

    if len(data) > 1:
        if data[1].columns[0].name == 'BMAJ':
            header["BMAJ"] = data[1].data["BMAJ"].mean()*arcsec * 180./numpy.pi
            header["BMIN"] = data[1].data["BMIN"].mean()*arcsec * 180./numpy.pi
            header["BPA"] = data[1].data["BPA"].mean()

    # Turn off the WCS warnings that come from the PCX_Y values because they
    # are annoying...
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=AstropyWarning)

        w = wcs.WCS(header)
        w = w.dropaxis(2)
        try:
            w = w.dropaxis(3)
        except:
            pass

    #x, y = numpy.meshgrid(numpy.linspace(0,nx-1,nx), numpy.linspace(0,ny-1,ny))
    x, y = None, None

    if header["CTYPE3"] in ["VELOCITY","VRAD"]:
        v0 = data[0].header["CRVAL3"]
        dv = data[0].header["CDELT3"]
        n0 = data[0].header["CRPIX3"]
        velocity = (numpy.arange(nfreq)-(n0-1))*dv/1000.+v0/1000.

        nu0 = data[0].header["RESTFREQ"]
        freq = nu0 - velocity*1e5 * nu0 / c
    elif header["CTYPE3"] == "FREQ":
        nu0 = data[0].header["CRVAL3"]
        dnu = data[0].header["CDELT3"]
        n0 = data[0].header["CRPIX3"]
        freq = (numpy.arange(nfreq)-(n0-1))*dnu + nu0

        velocity = None

    data.close()
 
    return Image(image, x=x, y=y, header=header, wcs=w, velocity=velocity, \
            freq=freq)



import numpy

def linear_regression(_x, _y, sigma_y=None):

    x = numpy.mat(_x).T
    y = numpy.mat(_y).T

    A = numpy.concatenate((x, numpy.ones(x.shape)), axis=1)

    if type(sigma_y) != type(None):
        C = numpy.diag(sigma_y**2)
    else:
        C = numpy.diag(numpy.ones(_y.shape))

    Xa = numpy.linalg.inv(A.T * numpy.linalg.inv(C) * A) * \
            (A.T * numpy.linalg.inv(C) * y)

    Cov = numpy.linalg.inv(A.T * numpy.linalg.inv(C) * A)

    return Xa[0,0], Xa[1,0], Cov[0,0]**0.5, Cov[1,1]**0.5



from ..constants.physics import c
from ..constants.math import pi
from ..constants.astronomy import Jy
from numpy import arange,ones,concatenate,sqrt,exp,where,array,mat,sin,log
from numpy import abs as absv
from scipy.optimize import leastsq
import matplotlib.pyplot as plt

def line_flux(data,lines,nleft=8,nright=8,plotout=None,quiet=False, \
    fixed_width=False,fringing=False):
    
    B = 1.0/1500
    
    lines = array(lines)
    
    wave = data.wave
    flux = data.flux
    unc = data.unc
    
    nlines = lines.size
    ind = arange(nlines)
    
    left = where(absv(wave-lines[0]) == absv(wave-lines[0]).min())
    right = where(absv(wave-lines[lines.size-1]) == absv(wave- \
        lines[lines.size-1]).min())
    
    prange = arange(right[0]-left[0]+nleft+nright+1)-nleft+left[0]
    
    wave_fit = wave[prange]
    flux_fit = flux[prange]
    unc_fit = unc[prange]
    
    flux_in = ones(lines.size)
    #sigma_in = lines*B
    fwhm_in = lines/(888.488-9.553*lines)
    sigma_in = fwhm_in/(2*sqrt(2*log(2)))
    
    omega_in = array([2*pi/(4*sigma_in[0])])
    amp_in = array([0.1])
    phi_in = array([pi/2])
    
    slope = array([(flux_fit[flux_fit.size-1]-flux_fit[0])/(wave_fit.max()- \
        wave_fit.min())])
    yint = array([flux_fit[0]-slope[0]*wave_fit.min()])
    
    if fringing:
        A = concatenate((flux_in,lines,sigma_in,yint,slope, \
                amp_in,omega_in,phi_in))
    else:
        A = concatenate((flux_in,lines,sigma_in,yint,slope))
    
    parinfo = []
    for i in arange(A.size):
        parinfo.append({"limited":[False,False], "limits":[0.0,0.0], \
            "mpside":2, "fixed":False})    
    
    for i in arange(lines.size):
        parinfo[i]["limited"] = [True,False]
        parinfo[i]["limits"] = [0.001,0.0]
        parinfo[lines.size+i]["limited"] = [True,True]
        parinfo[lines.size+i]["limits"] = [lines[i]-0.025,lines[i]+0.025]
        parinfo[2*lines.size+i]["limited"] = [True,False]
        parinfo[2*lines.size+i]["limits"] = [1.0e-3,0.0]
        if fixed_width:
            parinfo[2*lines.size+i]["fixed"] = True
    
    if fringing:
        parinfo[-1]["limited"] = [True,True]
        parinfo[-1]["limits"] = [0.0,2*pi]
    
    fa = {"x":wave_fit, "y":flux_fit, "err":unc_fit, "fringing":fringing}
    mfit = mpfit(gauss,xall=A,functkw=fa,parinfo=parinfo,quiet=1)
    
    A=mfit.params
    
    fit = gauss(A,x=wave_fit,y=flux_fit,err=unc_fit,fringing=fringing)[1]* \
            unc_fit*(-1)+flux_fit
    chisq = ((fit - flux_fit)**2/unc_fit**2).sum()/(wave_fit.size-A.size)
    
    # Calculate the flux and uncertainty.
    
    F = sqrt(2*pi)*A[ind]*Jy*(c*A[ind+2*nlines]*1.0e-4)/(A[ind+nlines]* \
        1.0e-4)**2/1.0e7
    
    deltaF = ones(nlines)
    for i in arange(nlines):
        deltaF[i] = sqrt(((unc_fit*Jy*c*B/(wave_fit*1.0e-4)*exp(-1.0*( \
            wave_fit-A[i+nlines])**2/(2*A[i+2*nlines]**2)))**2).sum())/1.0e7
    
    # Output the results.
    
    Results = concatenate((array(mat(A[ind+nlines]).T),array(mat(F).T), \
        array(mat(deltaF).T),array(mat(A[ind+2*nlines]*2*sqrt(2*log(2))).T)), \
        axis=1)
    
    if quiet == False:
        print("")
        for i in arange(lines.size):
            print("  {0:>6.3f}  {1:>9.3e}  {2:>9.3e}  {3:>6.4f}".format( \
                Results[i,0],Results[i,1],Results[i,2],Results[i,3]))
        print("")
        print("Reduced chi-squared of the fit: ",chisq)
        print("")
    
    # Plot the results.
    
    if (plotout != None) or (quiet == False):
        plt.errorbar(wave_fit,flux_fit,fmt="b",yerr=unc_fit)
        plt.plot(wave_fit,fit,"r")
        plt.xlabel("$\lambda$ [$\mu$"+"m]")
        plt.ylabel(r"F$_{\nu}$ [Jy]")
        if plotout != None:
            plt.savefig(plotout)
        elif quiet == False:
            plt.show()
    
    plt.clf()
    
    return Results, chisq

def gauss(p, fjac=None, x=None, y=None, err=None, fringing=False):
    
    if fringing:
        model = p[p.size-5]+p[p.size-4]*x+ \
                p[p.size-3]*sin(p[p.size-2]*x+p[p.size-1])
    
        for i in arange(p.size/3-1):
            model += p[i]*exp(-1*(x-p[p.size/3-1+i])**2/ \
                        (2*p[2*(p.size/3-1)+i]**2))
    else:
        model = p[p.size-2]+p[p.size-1]*x

        for i in arange(p.size/3):
            model += p[i]*exp(-1*(x-p[p.size/3+i])**2/ \
                        (2*p[2*(p.size/3)+i]**2))
    
    status = 0
    
    return [status, (y-model)/err]



from ..modeling import run_disk_model, run_flared_model
import numpy

# Define a likelihood function.

def lnlike(params, visibilities, images, spectra, parameters, plot, \
        model="disk", ncpus=1, ncpus_highmass=1, with_hyperion=False, \
        timelimit=3600, source="ObjName", nice=19, verbose=False, \
        ftcode="galario"):

    if model == "disk":
        m = run_disk_model(visibilities, images, spectra, params, \
                parameters, plot, ncpus=ncpus, ncpus_highmass=ncpus_highmass, \
                with_hyperion=with_hyperion, timelimit=timelimit, \
                source=source, nice=nice, verbose=verbose, ftcode=ftcode)
    elif model == "flared":
        m = run_flared_model(visibilities, params, parameters, plot, \
                ncpus=ncpus, source=source, nice=nice, ftcode=ftcode)

    # Catch whether the model timed out.

    if m == 0.:
        return -numpy.inf

    # A list to put all of the chisq into.

    chisq = []

    # Calculate the chisq for the visibilities.

    for j in range(len(visibilities["file"])):
        good = visibilities["data"][j].weights > 0

        chisq.append(-0.5*numpy.sum((visibilities["data"][j].real - \
                m.visibilities[visibilities["lam"][j]].real)**2 * \
                visibilities["data"][j].weights) - \
                numpy.sum(numpy.log(visibilities["data"][j].weights[good]/ \
                (2*numpy.pi))) + \
                -0.5*numpy.sum((visibilities["data"][j].imag - \
                m.visibilities[visibilities["lam"][j]].imag)**2 * \
                visibilities["data"][j].weights) - \
                numpy.sum(numpy.log(visibilities["data"][j].weights[good]/ \
                (2*numpy.pi))))


    # Calculate the chisq for all of the images.

    for j in range(len(images["file"])):
        chisq.append(-0.5 * (numpy.sum((images["data"][j].image - \
                m.images[images["lam"][j]].image)**2 / \
                images["data"][j].unc**2)))

    # Calculate the chisq for the SED.

    if "total" in spectra:
        chisq.append(-0.5 * (numpy.sum((spectra["total"].flux - \
                m.spectra["SED"].flux)**2 / spectra["total"].unc**2)))

    # Return the sum of the chisq.

    return numpy.array(chisq).sum()

# Define a prior function.

def lnprior(params, parameters, priors, visibilities):
    for key in parameters:
        if not parameters[key]["fixed"]:
            if parameters[key]["limits"][0] <= params[key] <= \
                    parameters[key]["limits"][1]:
                pass
            else:
                return -numpy.inf

    # Make sure that the radii are correct.

    if "logR_in" in params:
        R_in = 10.**params["logR_in"]
    else:
        R_in = 10.**parameters["logR_in"]["value"]

    if "logR_disk" in params:
        R_disk = 10.**params["logR_disk"]
    else:
        R_disk = 10.**parameters["logR_disk"]["value"]

    if "logR_env" in params:
        R_env = 10.**params["logR_env"]
    else:
        R_env = 10.**parameters["logR_env"]["value"]

    if R_in <= R_disk <= R_env:
        pass
    else:
        return -numpy.inf

    # Check that we aren't allowing any absurdly dense models.

    if "logR_env" in params and "logM_env" in params:
        if params["logR_env"] < 0.5 * params["logM_env"] + 4.:
            return -numpy.inf
        else:
            pass

    # Check that the cavity actually falls within the disk.

    if not parameters["logR_cav"]["fixed"]:
        if R_in <= 10.**params["logR_cav"] <= R_disk:
            pass
        else:
            return -numpy.inf

    # Check that the midplane temperature is below the atmosphere temperature.

    if ("logTmid0" in params) or ("logTmid0" in parameters):
        if "logTmid0" in params:
            Tmid0 = 10.**params["logTmid0"]
        else:
            Tmid0 = 10.**parameters["logTmid0"]["value"]

        if "logTatm0" in params:
            Tatm0 = 10.**params["logTatm0"]
        else:
            Tatm0 = 10.**parameters["logTatm0"]["value"]

        if Tmid0 < Tatm0:
            pass
        else:
            return -numpy.inf

    # Check that the gap is reasonable.

    if not parameters["logR_gap1"]["fixed"]:
        if R_in <= 10.**params["logR_gap1"] - params["w_gap1"]/2:
            pass
        else:
            return -numpy.inf

    # Everything was correct, so continue on.

    lnprior = 0.0

    # Add in the priors.

    for i in range(len(visibilities["file"])):
        if (not parameters["flux_unc{0:d}".format(i+1)]["fixed"]) and \
                (parameters["flux_unc{0:d}".format(i+1)]["prior"] == \
                "gaussian"):
            lnprior += -0.5 * (params["flux_unc{0:d}".format(i+1)] - \
                    parameters["flux_unc{0:d}".format(i+1)]["value"])**2 / \
                    parameters["flux_unc{0:d}".format(i+1)]["sigma"]**2

    # The prior on parallax (distance).

    if (not parameters["dpc"]["fixed"]) and ("parallax" in priors):
        parallax_mas = 1. / params["dpc"] * 1000

        lnprior += -0.5 * (parallax_mas - priors["parallax"]["value"])**2 / \
                priors["parallax"]["sigma"]**2
    elif (not parameters["dpc"]["fixed"]) and ("dpc" in priors):
        lnprior += -0.5 * (params["dpc"] - priors["dpc"]["value"])**2 / \
                priors["dpc"]["sigma"]**2

    # A prior on stellar mass from the IMF.

    if (not parameters["logM_star"]["fixed"]) and ("Mstar" in priors):
        Mstar = 10.**params["logM_star"]

        if priors["Mstar"]["value"] == "chabrier":
            if Mstar <= 1.:
                lnprior += numpy.log(0.158 * 1./(numpy.log(10.) * Mstar) * \
                        numpy.exp(-(numpy.log10(Mstar) - numpy.log10(0.08))**2/\
                        (2*0.69**2)))
            else:
                lnprior += numpy.log(4.43e-2 * Mstar**-1.3 * \
                        1./(numpy.log(10.) * Mstar))

    # Return

    return lnprior

# Define a probability function.

def lnprob(p, visibilities, images, spectra, parameters, priors, plot, \
        model="disk", ncpus=1, ncpus_highmass=1, with_hyperion=False, \
        timelimit=3600, source="ObjName", nice=19, verbose=False, \
        ftcode="galario"):

    keys = []
    for key in sorted(parameters.keys()):
        if not parameters[key]["fixed"]:
            keys.append(key)

    params = dict(zip(keys, p))

    lp = lnprior(params, parameters, priors, visibilities)

    if not numpy.isfinite(lp):
        return -numpy.inf

    return lp + lnlike(params, visibilities, images, spectra, parameters, \
            plot, model=model, ncpus=ncpus, ncpus_highmass=ncpus_highmass, \
            with_hyperion=with_hyperion, timelimit=timelimit, source=source, \
            nice=nice, verbose=verbose, ftcode=ftcode)



import numpy
cimport numpy
import h5py
import astropy
cimport cython
import time
from libc.math cimport pi

@cython.auto_pickle(True)
cdef class VisibilitiesObject:
    cdef public numpy.ndarray u, v, freq, real, imag, weights, \
            uvdist, amp, phase
    cdef public numpy.ndarray baseline
    cdef public str array_name

    def __init__(self, numpy.ndarray[double, ndim=1] u=None, \
            numpy.ndarray[double, ndim=1] v=None, \
            numpy.ndarray[double, ndim=1] freq=None, \
            numpy.ndarray[double, ndim=2] real=None, \
            numpy.ndarray[double, ndim=2] imag=None, \
            numpy.ndarray[double, ndim=2] weights=None, \
            baseline=None, array_name="CARMA"):

        if (type(u) != type(None)) and (type(v) != type(None)):
            self.u = u
            self.v = v
            self.uvdist = numpy.sqrt(u**2 + v**2)

        if (type(freq) != type(None)):
            self.freq = freq

        if (type(real) != type(None)) and (type(imag) != type(None)):
            self.real = real
            self.imag = imag
            self.amp = numpy.sqrt(real**2 + imag**2)
            self.phase = numpy.arctan2(imag, real)

            if type(weights) != type(None):
                self.weights = weights
            else:
                self.weights = numpy.ones((self.real.shape[0],self.real.shape[1]))

        self.baseline = baseline
        self.array_name = array_name

    def __reduce__(self):
        return (rebuild, (self.u, self.v, self.freq, self.real, self.imag, \
                self.weights, self.baseline, self.array_name))

def rebuild(u, v, freq, real, imag, weights, baseline, array_name):
    return VisibilitiesObject(u, v, freq, real, imag, weights, baseline, \
            array_name)

class Visibilities(VisibilitiesObject):

    def get_baselines(self, num):
        incl = self.baseline == num

        return Visibilities(self.u[incl], self.v[incl], self.freq, \
                self.real[incl,:], self.imag[incl,:], \
                self.weights[incl,:], baseline=self.baseline[incl])

    def set_header(self, header):
        self.header = header

    def asFITS(self):
        hdulist = astropy.fits.HDUList([])

        nvis = self.u.size
        hdu = astropy.fits.PrimaryHDU(numpy.concatenate((\
                self.u.reshape((1,nvis)), self.v.reshape((1,nvis)), \
                self.real.reshape((1,nvis)), self.imag.reshape((1,nvis)), \
                self.weights.reshape((1,nvis))), axis=0))

        if type(self.header) != type(None):
            hdu.header = self.header

        hdulist.append(hdu)

        return hdulist

    """
    def __reduce__(self):
        return (self.rebuild, (self.u, self.v, self.freq, self.real, self.imag,\
                self.weights))

    def rebuild(u, v, freq, real, imag, weights):
        return Visibilities(u, v, freq, real, imag, weights)
    """

    def read(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        u, v, freq, real, imag, weights = None, None, None, None, None, None

        if ('u' in f) and ('v' in f):
            u = f['u'][...].astype(numpy.double)
            v = f['v'][...].astype(numpy.double)
        if ('freq' in f):
            freq = f['freq'][...].astype(numpy.double)
        if ('real' in f) and ('imag' in f):
            real = f['real'][...].astype(numpy.double)
            imag = f['imag'][...].astype(numpy.double)

            if ('weights' in f):
                weights = f['weights'][...].astype(numpy.double)
            else:
                weights = numpy.ones(real.shape)

        self.__init__(u, v, freq, real, imag, weights)

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        if (type(self.u) != type(None)) and (type(self.v) != type(None)):
            u_dset = f.create_dataset("u", self.u.shape, dtype='float64')
            u_dset[...] = self.u
            v_dset = f.create_dataset("v", self.v.shape, dtype='float64')
            v_dset[...] = self.v
        if (type(self.freq) != type(None)):
            freq_dset = f.create_dataset("freq", self.freq.shape, \
                    dtype='float64')
            freq_dset[...] = self.freq
        if (type(self.real) != type(None)) and (type(self.imag) != type(None)):
            real_dset = f.create_dataset("real", self.real.shape, \
                    dtype='float64')
            real_dset[...] = self.real
            imag_dset = f.create_dataset("imag", self.imag.shape, \
                    dtype='float64')
            imag_dset[...] = self.imag

            if type(self.weights) != type(None):
                if numpy.product(self.weights == numpy.ones(self.real.shape)) \
                        == 0:
                    weights_dset = f.create_dataset("weights", \
                            self.weights.shape, dtype='float64')
                    weights_dset[...] = self.weights

        if (usefile == None):
            f.close()

@cython.boundscheck(False)
def average(data, gridsize=256, binsize=None, radial=False, log=False, \
        logmin=None, logmax=None, mfs=False, mode="continuum"):

    cdef numpy.ndarray[double, ndim=1] u, v, freq, uvdist
    cdef numpy.ndarray[double, ndim=2] real, imag, weights
    cdef numpy.ndarray[double, ndim=3] new_real, new_imag, new_weights
    cdef numpy.ndarray[unsigned int, ndim=1] i, j
    cdef unsigned int k, n
    
    if mfs:
        vis = freqcorrect(data)
        u = vis.u
        v = vis.v
        uvdist = vis.uvdist
        freq = vis.freq
        real = vis.real
        imag = vis.imag
        weights = vis.weights
    else:
        u = data.u.copy()
        v = data.v.copy()
        uvdist = data.uvdist.copy()
        freq = data.freq.copy()
        real = data.real.copy()
        imag = data.imag.copy()
        weights = data.weights
    
    # Set the weights equal to 0 when the point is flagged (i.e. weight < 0)
    weights = numpy.where(weights < 0,0.0,weights)
    # Set the weights equal to 0 when the real and imaginary parts are both 0
    weights[(real == 0) & (imag == 0)] = 0.0
    
    good_data = uvdist != 0.0
    u = u[good_data]
    v = v[good_data]
    uvdist = uvdist[good_data]
    real = real[good_data,:]
    imag = imag[good_data,:]
    weights = weights[good_data,:]

    # Set some parameter numbers for future use.

    cdef int nuv = u.size
    cdef int nfreq = freq.size
    
    cdef int nchannels
    if mode == "continuum":
        nchannels = 1
    elif mode == "spectralline":
        nchannels = freq.size

    # Average over the U-V plane by creating bins to average over.
    
    if radial:
        if log:
            temp = numpy.linspace(numpy.log10(logmin), numpy.log10(logmax), \
                    gridsize+1)
            new_u = 10**((temp[1:] + temp[0:-1])/2)
        else:
            new_u = numpy.linspace(binsize/2,(gridsize-0.5)*binsize,gridsize)
        new_u = new_u.reshape((1,gridsize))
        new_v = numpy.zeros((1,gridsize))
        new_real = numpy.zeros((1,gridsize,nchannels))
        new_imag = numpy.zeros((1,gridsize,nchannels))
        new_weights = numpy.zeros((1,gridsize,nchannels))

        if log:
            dtemp = temp[1] - temp[0]
            i = numpy.round((numpy.log10(uvdist)- numpy.log10(logmin))/dtemp - \
                    0.5).astype(numpy.uint32)
            j = numpy.zeros(uvdist.size).astype(numpy.uint32)
        else:
            i = numpy.round(uvdist/binsize).astype(numpy.uint32)
            j = numpy.zeros(uvdist.size).astype(numpy.uint32)
    else:
        if gridsize%2 == 0:
            uu = numpy.linspace(-gridsize*binsize/2, (gridsize/2-1)*binsize, \
                    gridsize)
            vv = numpy.linspace(-gridsize*binsize/2, (gridsize/2-1)*binsize, \
                    gridsize)
        else:
            uu = numpy.linspace(-(gridsize-1)*binsize/2, \
                    (gridsize-1)*binsize/2, gridsize)
            vv = numpy.linspace(-(gridsize-1)*binsize/2, \
                    (gridsize-1)*binsize/2, gridsize)

        #new_u, new_v = numpy.meshgrid(uu, vv)
        new_u = numpy.zeros((gridsize,gridsize,nchannels))
        new_v = numpy.zeros((gridsize,gridsize,nchannels))
        new_real = numpy.zeros((gridsize,gridsize,nchannels))
        new_imag = numpy.zeros((gridsize,gridsize,nchannels))
        new_weights = numpy.zeros((gridsize,gridsize,nchannels))

        if gridsize%2 == 0:
            i = numpy.round(u/binsize+gridsize/2.).astype(numpy.uint32)
            j = numpy.round(v/binsize+gridsize/2.).astype(numpy.uint32)
        else:
            i = numpy.round(u/binsize+(gridsize-1)/2.).astype(numpy.uint32)
            j = numpy.round(v/binsize+(gridsize-1)/2.).astype(numpy.uint32)

    good_i = numpy.logical_and(i >= 0, i < gridsize)
    good_j = numpy.logical_and(j >= 0, j < gridsize)
    good = numpy.logical_and(good_i, good_j)
    if good.sum() < good.size:
        print("WARNING: uv.grid was supplied with a gridsize and binsize that do not cover the full range of the input data in the uv-plane and is cutting baselines that are outside of this grid. Make sure to check your results carefully.")

    u = u[good]
    v = v[good]
    real = real[good,:]
    imag = imag[good,:]
    weights = weights[good,:]

    i = i[good]
    j = j[good]

    nuv = u.size

    for k in range(nuv):
        for n in range(nfreq):
            if mode == "continuum":
                if not radial:
                    new_u[j[k],i[k],0] += u[k]*weights[k,n]
                    new_v[j[k],i[k],0] += v[k]*weights[k,n]
                new_real[j[k],i[k],0] += real[k,n]*weights[k,n]
                new_imag[j[k],i[k],0] += imag[k,n]*weights[k,n]
                new_weights[j[k],i[k],0] += weights[k,n]
            elif mode == "spectralline":
                if not radial:
                    new_u[j[k],i[k],n] += u[k]*weights[k,n]
                    new_v[j[k],i[k],n] += v[k]*weights[k,n]
                new_real[j[k],i[k],n] += real[k,n]*weights[k,n]
                new_imag[j[k],i[k],n] += imag[k,n]*weights[k,n]
                new_weights[j[k],i[k],n] += weights[k,n]

    good_data = new_weights != 0.0
    new_real[good_data] = new_real[good_data] / new_weights[good_data]
    new_imag[good_data] = new_imag[good_data] / new_weights[good_data]
    if not radial:
        new_u[good_data] = new_u[good_data] / new_weights[good_data]
        new_v[good_data] = new_v[good_data] / new_weights[good_data]

    good_data = numpy.any(good_data, axis=2)
    if not radial:
        new_u = (new_u*new_weights).sum(axis=2)[good_data] / \
                new_weights.sum(axis=2)[good_data]
        new_v = (new_v*new_weights).sum(axis=2)[good_data] / \
                new_weights.sum(axis=2)[good_data]
    else:
        new_u = new_u[good_data]
        new_v = new_v[good_data]

    good_data = numpy.dstack([good_data for m in range(nchannels)])

    if mode == "continuum":
        freq = numpy.array([data.freq.sum()/data.freq.size])

    return Visibilities(new_u, new_v, freq, \
            new_real[good_data].reshape((new_u.size,nchannels)),\
            new_imag[good_data].reshape((new_u.size,nchannels)), \
            new_weights[good_data].reshape((new_u.size,nchannels)))

@cython.boundscheck(False)
def grid(data, gridsize=256, binsize=2000.0, convolution="pillbox", \
        mfs=False, channel=None, imaging=False, weighting="natural", \
        robust=2, npixels=0, mode="continuum"):
    
    cdef numpy.ndarray[double, ndim=1] u, v, freq
    cdef numpy.ndarray[double, ndim=2] real, imag, weights, new_u, new_v
    cdef numpy.ndarray[double, ndim=3] new_real, new_imag, new_weights
    cdef numpy.ndarray[double, ndim=3] binned_weights
    cdef numpy.ndarray[unsigned int, ndim=2] i, j
    cdef unsigned int k, l, m, n, ninclude_min, ninclude_max, lmin, lmax, \
            mmin, mmax, ll, mm, ninclude, npix
    cdef double mean_freq, inv_freq

    if mfs:
        vis = freqcorrect(data)
        u = vis.u
        v = vis.v
        freq = vis.freq
        real = vis.real
        imag = vis.imag
        weights = vis.weights
    else:
        u = data.u
        v = data.v
        if channel != None:
            freq = numpy.array([data.freq[channel]])
            real = data.real[:,channel].reshape((data.real.shape[0],1))
            imag = data.imag[:,channel].reshape((data.real.shape[0],1))
            weights = data.weights[:,channel]. \
                    reshape((data.real.shape[0],1))
        else:
            freq = data.freq
            real = data.real
            imag = data.imag
            weights = data.weights
    
    # Set the weights equal to 0 when the point is flagged (i.e. weight < 0)
    weights = numpy.where(weights < 0, 0.0, weights)
    # Set the weights equal to 0 when the real and imaginary parts are both 0
    weights[(real==0) & (imag==0)] = 0.0

    # Set some parameter numbers for future use.
    
    cdef double convolve
    cdef double inv_binsize = 1. / binsize
    cdef int nuv = u.size
    cdef int nfreq = freq.size

    cdef int nchannels
    if mode == "continuum":
        nchannels = 1
    elif mode == "spectralline":
        nchannels = freq.size

    # Average over the U-V plane by creating bins to average over.
    
    if gridsize%2 == 0:
        uu = numpy.linspace(-gridsize*binsize/2, (gridsize/2-1)*binsize, \
                gridsize)
        vv = numpy.linspace(-gridsize*binsize/2, (gridsize/2-1)*binsize, \
                gridsize)
    else:
        uu = numpy.linspace(-(gridsize-1)*binsize/2, (gridsize-1)*binsize/2, \
                gridsize)
        vv = numpy.linspace(-(gridsize-1)*binsize/2, (gridsize-1)*binsize/2, \
                gridsize)

    new_u, new_v = numpy.meshgrid(uu, vv)
    new_real = numpy.zeros((gridsize,gridsize,nchannels))
    new_imag = numpy.zeros((gridsize,gridsize,nchannels))
    new_weights = numpy.zeros((gridsize,gridsize,nchannels))

    # Get the indices for binning.

    i = numpy.zeros((nuv, nfreq), dtype=numpy.uint32)
    j = numpy.zeros((nuv, nfreq), dtype=numpy.uint32)

    mean_freq = numpy.mean(freq)
    inv_freq = 1./mean_freq

    if gridsize%2 == 0:
        i = numpy.array(u.reshape((u.size,1))*freq*inv_freq/binsize+\
                gridsize/2., dtype=numpy.uint32)
        j = numpy.array(v.reshape((v.size,1))*freq*inv_freq/binsize+\
                gridsize/2., dtype=numpy.uint32)
    else:
        i = numpy.array(u.reshape((u.size,1))*freq*inv_freq/binsize+ \
                    (gridsize-1)/2., dtype=numpy.uint32)
        j = numpy.array(v.reshape((v.size,1))*freq*inv_freq/binsize+ \
                    (gridsize-1)/2., dtype=numpy.uint32)
    
    if convolution == "pillbox":
        convolve_func = ones
        ninclude = 3
    elif convolution == "expsinc":
        convolve_func = exp_sinc
        ninclude = 6

    if ninclude%2 == 0:
        ninclude_min = numpy.uint32(ninclude*0.5-1)
        ninclude_max = numpy.uint32(ninclude*0.5)
    else:
        ninclude_min = numpy.uint32((ninclude-1)*0.5)
        ninclude_max = numpy.uint32((ninclude-1)*0.5)

    # Check whether any data falls outside of the grid, and exclude.

    good_i = numpy.logical_and(i >= 0, i < gridsize)
    good_j = numpy.logical_and(j >= 0, j < gridsize)
    good = numpy.logical_and(good_i, good_j)
    if good.sum() < good.size:
        print("WARNING: uv.grid was supplied with a gridsize and binsize that do not cover the full range of the input data in the uv-plane and is cutting baselines that are outside of this grid. Make sure to check your results carefully.")

    # If we are using a non-uniform weighting scheme, adjust the data weights.

    if weighting in ["uniform","superuniform","robust"]:
        binned_weights = numpy.ones((gridsize,gridsize,nchannels))

        npix = npixels

        if weighting == "superuniform":
            npix = 3

        for k in range(nuv):
            for n in range(nfreq):
                if not good[k,n]:
                    continue

                if npix > j[k,n]:
                    lmin = 0
                else:
                    lmin = j[k,n] - npix

                lmax = int_min(j[k,n]+npix+1, gridsize)

                if npix > i[k,n]:
                    mmin = 0
                else:
                    mmin = i[k,n] - npix

                mmax = int_min(i[k,n]+npix+1, gridsize)

                for l in range(lmin, lmax):
                    for m in range(mmin, mmax):
                        if mode == "continuum":
                            binned_weights[l,m,0] += weights[k,n]
                        elif mode == "spectralline":
                            binned_weights[l,m,n] += weights[k,n]

        if weighting in ["uniform","superuniform"]:
            for k in range(nuv):
                for n in range(nfreq):
                    if not good[k,n]:
                        continue

                    l = j[k,n]
                    m = i[k,n]

                    weights[k,n] /= binned_weights[l,m,n]
        elif weighting == "robust":
            f2 = (5*10**(-robust))**2 / \
                    ((binned_weights**2).sum(axis=(0,1)) / weights.sum(axis=0))

            for k in range(nuv):
                for n in range(nfreq):
                    if not good[k,n]:
                        continue

                    l = j[k,n]
                    m = i[k,n]

                    weights[k,n] /= (1 + f2[n] * binned_weights[l,m,n])

    # Now actually go through and calculate the new visibilities.

    for k in range(nuv):
        for n in range(nfreq):
            if not good[k,n]:
                continue

            if ninclude_min > j[k,n]:
                lmin = 0
            else:
                lmin = j[k,n] - ninclude_min

            lmax = int_min(j[k,n]+ninclude_max+1, gridsize)

            if ninclude_min > i[k,n]:
                mmin = 0
            else:
                mmin = i[k,n] - ninclude_min

            mmax = int_min(i[k,n]+ninclude_max+1, gridsize)

            for l in range(lmin, lmax):
                for m in range(mmin, mmax):
                    convolve = convolve_func( (u[k]*freq[n]*inv_freq-\
                            new_u[l,m])*inv_binsize, \
                            (v[k]*freq[n]*inv_freq - new_v[l,m]) * inv_binsize)

                    if mode == "continuum":
                        new_real[l,m,0] += real[k,n]*weights[k,n]*convolve
                        new_imag[l,m,0] += imag[k,n]*weights[k,n]*convolve
                        new_weights[l,m,0] += weights[k,n]*convolve
                    elif mode == "spectralline":
                        new_real[l,m,n] += real[k,n]*weights[k,n]*convolve
                        new_imag[l,m,n] += imag[k,n]*weights[k,n]*convolve
                        new_weights[l,m,n] += weights[k,n]*convolve

    # If we are making an image, normalize the weights.

    if imaging:
        for n in range(nchannels):
            new_real[:,:,n] /= new_weights[:,:,n].sum()
            new_imag[:,:,n] /= new_weights[:,:,n].sum()
            new_weights[:,:,n] /= new_weights[:,:,n].sum()
    else:
        good = new_weights > 0
        new_real[good] = new_real[good] / new_weights[good]
        new_imag[good] = new_imag[good] / new_weights[good]

    if mode == "continuum":
        freq = numpy.array([data.freq.sum()/data.freq.size])
    
    return Visibilities(new_u.reshape(gridsize**2), new_v.reshape(gridsize**2),\
            freq, new_real.reshape((gridsize**2,nchannels)), \
            new_imag.reshape((gridsize**2,nchannels)), \
            new_weights.reshape((gridsize**2,nchannels)))

cdef inline int int_max(int a, int b): return a if a >= b else b
cdef inline int int_min(int a, int b): return a if a <= b else b
cdef inline double int_abs(double a): return -a if a < 0 else a

cdef double sinc(double x):

    cdef double xp = x * pi

    return 1. - xp**2/6. + xp**4/120. - xp**6/5040. + xp**8/362880. - \
            xp**10/39916800. + xp**12/6227020800. - xp**14/1307674368000. + \
            xp**16/355687428096000.

cdef double exp(double x):

    return 1 + x + x**2/2. + x**3/6. + x**4/24. + x**5/120.

cdef double exp_sinc(double u, double v):
    
    cdef double inv_alpha1 = 1. / 1.55
    cdef double inv_alpha2 = 1. / 2.52
    cdef double norm = 2.350016262343186
    cdef int m = 6
    
    if (int_abs(u) >= m * 0.5) or (int_abs(v) >= m * 0.5):
        return 0.

    cdef double arr = sinc(u * inv_alpha1) * \
            sinc(v * inv_alpha1) * \
            exp(-1 * (u * inv_alpha2)**2) * \
            exp(-1 * (v * inv_alpha2)**2) / norm

    return arr

cdef double ones(double u, double v):
    
    cdef int m = 1

    if (int_abs(u) >= m * 0.5) or (int_abs(v) >= m * 0.5):
        return 0.

    cdef double arr = 1.0

    return arr

@cython.boundscheck(False)
def freqcorrect(data, freq=None):
    cdef numpy.ndarray[double, ndim=1] new_u, new_v, new_freq
    cdef numpy.ndarray[double, ndim=2] new_real, new_imag, new_weights

    if freq != None:
        new_freq = numpy.array([freq])
    else:
        new_freq = numpy.array([data.freq.mean()])

    cdef double inv_freq = 1./new_freq[0]
    cdef numpy.ndarray[double, ndim=1] scale = data.freq * inv_freq

    new_u = (data.u.reshape((data.u.size,1))*scale).reshape((data.real.size,))
    new_v = (data.v.reshape((data.v.size,1))*scale).reshape((data.real.size,))

    new_real = data.real.reshape((data.real.size,1))
    new_imag = data.imag.reshape((data.imag.size,1))
    new_weights = data.weights.reshape((data.weights.size,1))

    return Visibilities(new_u, new_v, new_freq, new_real, new_imag, \
            new_weights)

def chisq(data, model):
    chi_squared = chisq_calc(data.real, data.imag, data.weights, model.real, \
            model.imag, data.real.size)

    return chi_squared

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cdef float chisq_calc(numpy.ndarray[double, ndim=2] data_real, \
        numpy.ndarray[double, ndim=2] data_imag, \
        numpy.ndarray[double, ndim=2] data_weights, \
        numpy.ndarray[double, ndim=2] model_real, \
        numpy.ndarray[double, ndim=2] model_imag, int nuv):
    cdef double chisq = 0
    cdef double diff1, diff2
    cdef unsigned int i

    for i in range(nuv):
        diff1 = data_real[<unsigned int>i,0] - model_real[<unsigned int>i,0]
        diff2 = data_imag[<unsigned int>i,0] - model_imag[<unsigned int>i,0]
        chisq += (diff1*diff1 + diff2*diff2) * data_weights[<unsigned int>i,0]

    return chisq



#import matplotlib.pyplot as plt
from ..constants.astronomy import arcsec
import numpy
import scipy
import dynesty
import dynesty.utils as dyfunc
#import dynesty.plotting as dyplot
from .model import model

def fit_model(data, funct='point', nsteps=1e3, niter=3, max_size=numpy.inf, \
        xmax=10., ymax=10., step_size=0.1, min_separation=1., \
        primary_beam=None, image_rms=None):

    if type(funct) == str:
        funct = numpy.array([funct])
    elif type(funct) == list:
        funct = numpy.array(funct)
    elif type(funct) == numpy.ndarray:
        pass

    # Check if we need to adjust the data weights based on a supplied 
    # image_rms.

    if image_rms != None:
        sigma_tot = 1./numpy.sqrt(data.weights.sum())
        ratio = sigma_tot / image_rms

        data.weights *= ratio**2
    else:
        # If not, we could still use the image rms as a lower limit on flux
        # for the modeling.
        image_rms = 1./numpy.sqrt(data.weights.sum())

    # If the max_size == inf, calculate based on the data as dynesty needs a
    # finite range.

    if max_size == numpy.inf:
        max_size = 1./data.uvdist[data.uvdist > 0].min() / arcsec

    # Set up the data properly.

    x = data.u
    y = data.v
    z = numpy.concatenate((data.real, data.imag))[:,0]
    zerr = 1./numpy.sqrt(numpy.concatenate((data.weights,data.weights)))[:,0]

    # Set up some parameters for dynesty.

    periodic = []
    nparams = numpy.zeros(funct.size)
    for i in range(funct.size):
        if funct[i] == "point":
            nparams[i] = 3
        elif funct[i] == "gauss":
            nparams[i] = 6
        elif funct[i] == "circle":
            nparams[i] = 6
        elif funct[i] == "ring":
            nparams[i] = 7

        if funct[i] in ['gauss','circle','ring']:
            periodic.append(int(nparams.sum()-2))

    ndim = int(nparams.sum())
    nlive = int(4*xmax*ymax / step_size**2)

    xlim = [[-xmax, xmax] for i in range(funct.size)]
    ylim = [[-ymax, ymax] for i in range(funct.size)]

    # Set up the dynesty sampler.

    sampler = dynesty.NestedSampler(lnlike, ptform, ndim=ndim, nlive=250, \
            bound='multi', logl_args=(x, y, z, zerr, funct, nparams,\
            primary_beam, data.freq.mean()), ptform_args=(funct, nparams, \
            xlim, ylim, max_size, min_separation, image_rms), periodic=periodic)

    # Run the nested sampling.

    sampler.run_nested(dlogz=0.05)

    # Get the samples.

    res = sampler.results
    samples, weights = res.samples, numpy.exp(res.logwt - res.logz[-1])
    samples = dyfunc.resample_equal(samples, weights)

    # Make a traceplot.

    """
    fig, axes = dyplot.traceplot(res, show_titles=True, trace_cmap='viridis', \
            connect=True, connect_highlight=range(5))

    plt.show(fig)
    """

    # Correct values that are in log.

    ind = 0
    for i in range(funct.size):
        if funct[i] in ['gauss','circle','ring']:
            samples[:,int(ind+2)] = 10.**samples[:,int(ind+2)]

            if funct[i] in ['gauss','ring']:
                samples[:,int(ind+3)] = 10.**samples[:,int(ind+3)]

        samples[:,int(ind+nparams[i]-1)] = 10.**samples[:,int(ind+nparams[i]-1)]

        ind += nparams[i]

    # Get the best fit parameters.

    params = numpy.median(samples, axis=0)
    sigma = samples.std(axis=0)

    # Clip a few stragglers and re-calculate.

    return params, sigma, samples

# Define a likelihood function.

def lnlike(p, x, y, z, zerr, funct, nparams, primary_beam, freq):
    params = p.copy()

    # p values are supplied in log-space for many parameters, so correct those.
    ind = 0
    for i in range(funct.size):
        if funct[i] in ['gauss','circle','ring']:
            params[int(ind+2)] = 10.**params[int(ind+2)]

            if funct[i] in ['gauss','ring']:
                params[int(ind+3)] = 10.**params[int(ind+3)]

        params[int(ind+nparams[i]-1)] = 10.**params[int(ind+nparams[i]-1)]

        ind += nparams[i]

    # Generate the model.
    m = model(x, y, params, return_type="append", funct=funct, \
            primary_beam=primary_beam, freq=freq)

    # Calculate the chi^2 value.
    return -0.5*(numpy.sum((z - m)**2 / zerr**2))

# Define the prior transformation function.

def ptform(u, funct, nparams, xlim, ylim, max_size, min_separation, image_rms):
    p = numpy.array(u)

    # Log of the maximum size.
    logr_max = numpy.log10(max_size)

    # Use 3*rms as minimum flux value allowed.
    logF_min = numpy.log10(3. * image_rms)

    ind = 0
    x0, y0 = [], []
    for i in range(funct.size):
        # Always sample x from the full range.
        p[int(ind+0)] = (xlim[i][1] - xlim[i][0])*u[int(ind+0)] + xlim[i][0]

        # For y, we need to sample from everywhere but the y-values that are
        # too close to a previous source, for a given x.
        if i == 0:
            p[int(ind+1)] = (ylim[i][1] - ylim[i][0])*u[int(ind+1)] + \
                    ylim[i][0]
        else:
            # Only use the slow generation function if the x value of a previous
            # component is too close to the x-value of this component.
            if numpy.all(numpy.abs(p[int(ind+0)] - numpy.array(x0)) > \
                    min_separation):
                p[int(ind+1)] = (ylim[i][1] - ylim[i][0])*u[int(ind+1)] + \
                        ylim[i][0]
            else:
                p[int(ind+1)] = y_gen(ylim[i][0], ylim[i][1], p[int(ind+0)], \
                        x0, y0, min_separation).ppf(u[int(ind+1)])

        # Update our list of sources to watch out for.
        x0.append(p[int(ind+0)])
        y0.append(p[int(ind+1)])

        # Set ranges for models that have a size.
        if funct[i] in ['gauss','circle','ring']:
            p[int(ind+2)] = (logr_max - -3.)*u[int(ind+2)] - 3.

            if funct[i] in ['gauss','ring']:
                p[int(ind+3)] = (p[int(ind+2)] - -3.)*u[int(ind+3)] - 3.

            # Set the inclination for models that use it.
            if funct[i] == 'circle':
                p[int(ind+3)] = numpy.pi/2 * u[int(ind+3)]
            elif funct[i] == 'ring':
                p[int(ind+4)] = numpy.pi/2 * u[int(ind+4)]

            # Set the position angle.
            p[int(ind+nparams[i]-2)] = numpy.pi*u[int(ind+nparams[i]-2)]

        # Each source has to be less bright than the last.
        if i == 0:
            logF_max = 0.
        else:
            logF_max = p[int(ind-1)]

        p[int(ind+nparams[i]-1)] = (logF_max - logF_min)*\
                u[int(ind+nparams[i]-1)] + logF_min

        ind += nparams[i]

    return p

# Functions for helping to sample multiple systems.

def func(y, x, x0, y0, radius):
    if type(x0) == list:
        x0 = numpy.array(x0)
    if type(y0) == list:
        y0 = numpy.array(y0)

    return numpy.maximum(0, 1 - numpy.exp(-(((x - x0)**2 + (y - y0)**2) / \
            radius**2)**4).sum())

class y_gen(scipy.stats.rv_continuous):
    def __init__(self, a=None, b=None, x=None, x0=None, y0=None, radius=None):
        self.x = x
        self.x0 = x0
        self.y0 = y0
        self.radius = radius
        
        self.norm = scipy.integrate.quad(func, a, b, args=(x, x0, y0, \
                radius))[0]

        super().__init__(a=a, b=b)

    def _pdf(self, y):
        return func(y, self.x, self.x0, self.y0, self.radius) / self.norm



from numpy import sqrt, exp, pi

def ml(x, y, sigma_y, params, model, args=None, limits=None):

    if (type(limits) != type(None)):
        for i in range(params.size):
            if ((limits[i]["limited"][0] == True) and \
                    (params[i] < limits[i]["limits"][0])):
                ml = 0.0
                chisq = 1.0e300
                return ml, chisq
            elif ((limits[i]["limited"][1] == True) and \
                    (params[i] > limits[i]["limits"][1])):
                ml = 0.0
                chisq = 1.0e300
                return ml, chisq

    m = model(x, params, **args)

    mlarr = 1./sqrt(2*pi*sigma_y**2)*\
                exp(-(y-m)**2/(2*sigma_y**2)) 

    #ml = (mlarr**(1./(y.size-params.size))).prod()
    ml = 1.
    chisq = ((y-m)**2/sigma_y**2).sum()

    return ml, chisq



import matplotlib.pyplot as plt
import matplotlib.ticker as ticker

def plot_2D_visibilities(visibilities, model, parameters, params, index=0, \
        fig=None):

    # Generate a figure and axes if none is provided.

    if fig == None:
        fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(9.,4.))
    else:
        fig, ax = fig

    # Calculate the pixel range to show.

    ticks = visibilities["ticks"][index]

    xmin, xmax = int(round(visibilities["npix"][index]/2+ticks[0]/\
            (visibilities["binsize"][index]/1000))), \
            int(round(visibilities["npix"][index]/2+ticks[-1]/\
            (visibilities["binsize"][index]/1000)))
    ymin, ymax = int(round(visibilities["npix"][index]/2+ticks[0]/\
            (visibilities["binsize"][index]/1000))), \
            int(round(visibilities["npix"][index]/2+ticks[-1]/\
            (visibilities["binsize"][index]/1000)))

    # How to scale the real part.

    vmin = min(0, visibilities["data1d"][index].real.min())
    vmax = visibilities["data1d"][index].real.max()

    # Show the real component.

    ax[0].imshow(visibilities["data2d"][index].real.reshape(\
            (visibilities["npix"][index],visibilities["npix"][index]))\
            [xmin:xmax,xmin:xmax][:,::-1], origin="lower", \
            interpolation="nearest", vmin=vmin, vmax=vmax, cmap="jet")

    ax[0].contour(model.visibilities[visibilities["lam"][index]+"_2d"].real.\
            reshape((visibilities["npix"][index],visibilities["npix"][index]))\
            [xmin:xmax,xmin:xmax][:,::-1], cmap="jet")

    # How to scale the imaginary part.

    vmin = -visibilities["data1d"][index].real.max()
    vmax =  visibilities["data1d"][index].real.max()

    # Show the imaginary component.

    ax[1].imshow(visibilities["data2d"][index].imag.reshape(\
            (visibilities["npix"][index],visibilities["npix"][index]))\
            [xmin:xmax,xmin:xmax][:,::-1], origin="lower", \
            interpolation="nearest", vmin=vmin, vmax=vmax, cmap="jet")

    ax[1].contour(model.visibilities[visibilities["lam"][index]+"_2d"].imag.\
            reshape((visibilities["npix"][index],visibilities["npix"][index]))\
            [xmin:xmax,xmin:xmax][:,::-1], cmap="jet")

    # Adjust the axes ticks.

    transform1 = ticker.FuncFormatter(Transform(xmin, xmax, \
            visibilities["binsize"][index]/1000, '%.0f'))

    ax[0].set_xticks(visibilities["npix"][index]/2+ticks[1:-1]/\
            (visibilities["binsize"][index]/1000)-xmin)
    ax[0].set_yticks(visibilities["npix"][index]/2+ticks[1:-1]/\
            (visibilities["binsize"][index]/1000)-ymin)
    ax[0].get_xaxis().set_major_formatter(transform1)
    ax[0].get_yaxis().set_major_formatter(transform1)

    ax[1].set_xticks(visibilities["npix"][index]/2+ticks[1:-1]/\
            (visibilities["binsize"][index]/1000)-xmin)
    ax[1].set_yticks(visibilities["npix"][index]/2+ticks[1:-1]/\
            (visibilities["binsize"][index]/1000)-ymin)
    ax[1].get_xaxis().set_major_formatter(transform1)
    ax[1].get_yaxis().set_major_formatter(transform1)

    # Adjust the plot and save it.

    ax[0].set_xlabel("U [k$\lambda$]")
    ax[0].set_ylabel("V [k$\lambda$]")

    ax[1].set_xlabel("U [k$\lambda$]")
    ax[1].set_ylabel("V [k$\lambda$]")

    # Return the figure and axes.

    return fig, ax

# Define a useful class for plotting.

class Transform:
    def __init__(self, xmin, xmax, dx, fmt):
        self.xmin = xmin
        self.xmax = xmax
        self.dx = dx
        self.fmt = fmt

    def __call__(self, x, p):
        return self.fmt% ((x-(self.xmax-self.xmin+1)/2)*self.dx)



#!/usr/bin/env python3

from pdspy.utils.dynesty import lnlike, ptform
from matplotlib.backends.backend_pdf import PdfPages
import pdspy.plotting as plotting
import pdspy.modeling as modeling
import pdspy.utils as utils
import dynesty.plotting as dyplot
import dynesty.results as dyres
import dynesty.utils as dyfunc
import dynesty
import matplotlib.pyplot as plt
import astropy.stats
import schwimmbad
import argparse
import numpy
import sys
import os
import corner
from mpi4py import MPI

comm = MPI.COMM_WORLD

################################################################################
#
# Parse command line arguments.
#
################################################################################

parser = argparse.ArgumentParser()
parser.add_argument('-o', '--object')
parser.add_argument('-r', '--resume', action='store_true')
parser.add_argument('-p', '--resetprob', action='store_true')
parser.add_argument('-a', '--action', type=str, default="run")
parser.add_argument('-n', '--ncpus', type=int, default=1)
parser.add_argument('-e', '--withexptaper', action='store_true')
parser.add_argument('-v', '--plot_vis', action='store_true')
parser.add_argument('-c', '--withcontsub', action='store_true')
parser.add_argument('-b', '--trim', type=str, default="")
parser.add_argument('-i', '--nice', action='store_true')
parser.add_argument('-l', '--nicelevel', type=int, default=19)
parser.add_argument('-f', '--ftcode', type=str, default="galario")
args = parser.parse_args()

# Check whether we are using MPI.

withmpi = comm.Get_size() > 1

# Set the number of cpus to use.

ncpus = args.ncpus

# Set the nice level of the subprocesses.

if args.nice:
    nice = args.nicelevel
else:
    nice = None

# Get the source name and check that it has been set correctly.

source = args.object

if source == None:
    print("--object must be specified")
    sys.exit()

# Determine what action should be taken and set some variables appropriately.

if args.action not in ['run','plot']:
    print("Please select a valid action")
    sys.exit(0)

if args.action == 'plot':
    args.resume = True

# Get the items we want to trim.

if args.trim == "":
    trim = []
else:
    trim = args.trim.split(",")

################################################################################
#
# In case we are restarting this from the same job submission, delete any
# temporary directories associated with this run.
#
################################################################################

os.system("rm -r /tmp/temp_{0:s}_{1:d}".format(source, comm.Get_rank()))

################################################################################
#
# Set up a pool for parallel runs.
#
################################################################################

if args.action == "run":
    if withmpi:
        pool = schwimmbad.MPIPool()

        if not pool.is_master():
            pool.wait()
            sys.exit(0)
    else:
        pool = None
else:
    pool = None

################################################################################
#
# Read in the data.
#
################################################################################

# Import the configuration file information.

config = utils.load_config()

# Decide whether to use an exponentially tapered 

if args.withexptaper:
    config.parameters["disk_type"]["value"] = "exptaper"

# Decide whether to do continuum subtraction or not.

if args.withcontsub:
    config.parameters["docontsub"]["value"] = True

# Read in the data.

visibilities, images, spectra = utils.load_data(config, model="flared")

################################################################################
#
# Fit the model to the data.
#
################################################################################

# Set up the emcee run.

ndim = 0
periodic = []
keys = []

for key in sorted(config.parameters.keys()):
    if not config.parameters[key]["fixed"]:
        ndim += 1
        keys.append(key)

        if key == "pa":
            periodic.append(ndim-1)

# Make the labels nice with LaTeX.

labels = ["$"+key.replace("T0_env","T_0,env").replace("T0","T_0").\
        replace("Tatm0","T_atm,0").replace("Tmid0","T_mid,0").\
        replace("turb_env","turb,env").replace("in_env","in,env").\
        replace("_","_{").replace("log","\log ")+"}$" if key[0:3] == \
        "log" else "$"+key+"$" for key in keys]

# Set up the MCMC simulation.

if args.resume:
    """
    sampler = utils.dynesty.load_sampler("sampler.p", pool=pool)
    """
    sampler = dynesty.NestedSampler.restore("sampler.save", pool=pool)

    # For backwards compatibility
    sampler.loglikelihood.args = [visibilities, images, spectra, \
            config.parameters, False]
    sampler.loglikelihood.kwargs = {"model":"flared", \
            "ncpus":ncpus, "source":source, "nice":nice, "ftcode":args.ftcode}
    sampler.prior_transform.kwargs = {"model":"flared"}

    res = sampler.results
else:
    sampler = dynesty.NestedSampler(utils.dynesty.lnlike, utils.dynesty.ptform,\
            ndim, nlive=config.nlive_init, logl_args=(visibilities, images, \
            spectra, config.parameters, False), logl_kwargs={"model":"flared", \
            "ncpus":ncpus, "source":source, "nice":nice, \
            "ftcode":args.ftcode}, ptform_args=(\
            config.parameters, config.priors), ptform_kwargs={"model":\
            "flared"}, periodic=periodic, pool=pool, sample="rwalk", \
            walks=config.walks)

# Run a few burner steps.

if args.action == "run":
    add_live = False
    while not sampler.added_live:
        sampler.run_nested(dlogz=config.dlogz, checkpoint_file="sampler.save", \
                resume=args.resume, maxiter=1000, add_live=add_live, \
                checkpoint_every=1800)

        # Make plots of the status of the fit.

        if not add_live:
            sampler.add_final_live()

        res = sampler.results

        utils.dynesty.plot_status(res, ptform=sampler.prior_transform, \
                labels=labels, periodic=periodic)

        if not add_live:
            sampler._remove_live_points()

        # Manually calculate the stopping criterion to determine whether to 
        # add live points.

        logz_remain = numpy.max(sampler.live_logl) + \
                sampler.saved_run["logvol"][-1]
        delta_logz = numpy.logaddexp(sampler.saved_run["logz"][-1], \
                logz_remain) - sampler.saved_run["logz"][-1]

        add_live = delta_logz < config.dlogz

        # Once we've made it this far, we always want to resume.

        args.resume = True

# If we are just plotting, a few minor things to do.

elif args.action == "plot":
    # Add the final live points if needed and get the results.

    if not sampler.added_live:
        sampler.add_final_live()

        res = sampler.results
    else:
        res = sampler.results

    # Make the traceplots and the bound plots.

    utils.dynesty.plot_status(res, ptform=sampler.prior_transform, \
            labels=labels, periodic=periodic)

# Generate a plot of the weighted samples.

fig, ax = plt.subplots(ndim-1, ndim-1, figsize=(10,10))

dyplot.cornerpoints(res, cmap="plasma", kde=False, fig=(fig,ax), labels=labels)

fig.savefig("cornerpoints.png")

# Generate a corner plot from Dynesty.

fig, ax = plt.subplots(ndim, ndim, figsize=(15,15))

dyplot.cornerplot(res, color="blue", show_titles=True, max_n_ticks=3, \
        quantiles=None, fig=(fig, ax), labels=labels)

fig.savefig("cornerplot.png")

# Convert the results to a more traditional set of samples that you would
# get from an MCMC program.

samples, weights = res.samples, numpy.exp(res.logwt - res.logz[-1])

samples = dyfunc.resample_equal(samples, weights)

# Save pos, prob, chain.

numpy.save("samples.npy", samples)

# Make the cuts specified by the user.

for command in trim:
    command = command.split(" ")

    for i, key in enumerate(keys):
        if key == command[0]:
            if command[1] == '<':
                good = samples[:,i] > float(command[2])
            else:
                good = samples[:,i] < float(command[2])

            samples = samples[good,:]

# Now calculate the best fit parameters.

params = numpy.median(samples, axis=0)
sigma = astropy.stats.mad_std(samples, axis=0)

# Write out the results.

f = open("fit.txt", "w")
f.write("Best fit parameters:\n\n")
for j in range(len(keys)):
    f.write("{0:s} = {1:f} +/- {2:f}\n".format(keys[j], params[j], \
            sigma[j]))
f.write("\n")
f.close()

os.system("cat fit.txt")

# Plot histograms of the resulting parameters.

fig = corner.corner(samples, labels=labels, truths=params)

plt.savefig("fit.pdf")

# Make a dictionary of the best fit parameters.

params = dict(zip(keys, params))

############################################################################
#
# Plot the results.
#
############################################################################

# Create a high resolution model for averaging.

m = modeling.run_flared_model(visibilities, params, config.parameters, \
        plot=True, ncpus=ncpus, source=source, plot_vis=args.plot_vis, \
        nice=nice, ftcode=args.ftcode)

# Open up a pdf file to plot into.

pdf = PdfPages("model.pdf")

# Loop through the visibilities and plot.

for j in range(len(visibilities["file"])):
    # Plot the best fit model over the data.

    fig, ax = plt.subplots(nrows=visibilities["nrows"][j], \
            ncols=visibilities["ncols"][j], sharex=True, sharey=True)

    # Make a plot of the channel maps.

    plotting.plot_channel_maps(visibilities, m, config.parameters, params, \
            index=j, plot_vis=args.plot_vis, fig=(fig,ax))
    
    # Adjust the plot and save it.

    fig.set_size_inches((10,9))
    fig.subplots_adjust(left=0.08, right=0.98, top=0.98, bottom=0.07, \
            wspace=0.0,hspace=0.0)

    # Adjust the figure and save.

    pdf.savefig(fig)

    plt.clf()

# Close the pdf.

pdf.close()

# Now we can close the pool.

if args.action == "run":
    if withmpi:
        pool.close()



#!/usr/bin/env python3

import pdspy.modeling as modeling
import pdspy.plotting as plotting
import pdspy.utils as utils
import matplotlib.pyplot as plt
import astropy.stats
import schwimmbad
import argparse
import numpy
import sys
import os
import emcee
import corner
from mpi4py import MPI

comm = MPI.COMM_WORLD

################################################################################
#
# Parse command line arguments.
#
################################################################################

parser = argparse.ArgumentParser()
parser.add_argument('-o', '--object')
parser.add_argument('-r', '--resume', action='store_true')
parser.add_argument('-c', '--withhyperion', action='store_true')
parser.add_argument('-p', '--resetprob', action='store_true')
parser.add_argument('-a', '--action', type=str, default="run")
parser.add_argument('-n', '--ncpus', type=int, default=1)
parser.add_argument('-m', '--ncpus_highmass', type=int, default=8)
parser.add_argument('-e', '--withexptaper', action='store_true')
parser.add_argument('-t', '--timelimit', type=int, default=7200)
parser.add_argument('-b', '--trim', type=str, default="")
parser.add_argument('-s', '--SED', action='store_true')
parser.add_argument('-i', '--nice', action='store_true')
parser.add_argument('-l', '--nicelevel', type=int, default=19)
parser.add_argument('-v', '--verbose', action='store_true')
parser.add_argument('-f', '--ftcode', type=str, default="galario")
args = parser.parse_args()

# Check whether we are using MPI.

withmpi = comm.Get_size() > 1

# Set the number of cpus to use.

ncpus = args.ncpus
ncpus_highmass = args.ncpus_highmass

# Set the nice level of the subprocesses.

if args.nice:
    nice = args.nicelevel
else:
    nice = None

# Get the source name and check that it has been set correctly.

source = args.object

if source == None:
    print("--object must be specified")
    sys.exit()

# Determine what action should be taken and set some variables appropriately.

if args.action not in ['run','plot']:
    print("Please select a valid action")
    sys.exit(0)

if args.action == 'plot':
    args.resume = True

# Get the items we want to trim.

if args.trim == "":
    trim = []
else:
    trim = args.trim.split(",")

################################################################################
#
# In case we are restarting this from the same job submission, delete any
# temporary directories associated with this run.
#
################################################################################

os.system("rm -r /tmp/temp_{0:s}_{1:d}".format(source, comm.Get_rank()))

################################################################################
#
# Set up a pool for parallel runs.
#
################################################################################

if args.action == "run":
    if withmpi:
        pool = schwimmbad.MPIPool()

        if not pool.is_master():
            pool.wait()
            sys.exit(0)
    else:
        pool = None

################################################################################
#
# Read in the data.
#
################################################################################

# Import the configuration file information.

config = utils.load_config()

# Deprecated option to include an expontial taper on the disk.

if args.withexptaper:
    config.parameters["disk_type"]["value"] = "exptaper"

# Read in the data.

visibilities, images, spectra = utils.load_data(config)

################################################################################
#
# Fit the model to the data.
#
################################################################################

# Set up the emcee run.

ndim = 0
keys = []

for key in sorted(config.parameters.keys()):
    if not config.parameters[key]["fixed"]:
        ndim += 1
        keys.append(key)

# Make the labels nice with LaTeX.

labels = ["$"+key.replace("nu_ff","nu,ff").replace("_","_{").\
        replace("log","\log ")+"}$" if key[0:3] == "log" else \
        "$"+key.replace("h_large","h,large")+"$" for key in keys]

# If we are resuming an MCMC simulation, read in the necessary info, otherwise
# set up the info.

if args.resume:
    backend = emcee.backends.HDFBackend("results.hdf5")

    if args.resetprob:
        sample, _, _ = backend.get_last_sample()
    else:
        sample = backend.get_last_sample()
else:
    pos = []
    for j in range(config.nwalkers):
        pos.append(utils.propose_point_emcee(config.parameters))

    backend = emcee.backends.HDFBackend("results.hdf5")
    backend.reset(config.nwalkers, ndim)

    sample = emcee.State(pos)

# Set up the MCMC simulation.

if args.action == "run":
    sampler = emcee.EnsembleSampler(config.nwalkers, ndim, utils.emcee.lnprob, \
            args=(visibilities, images, spectra, config.parameters, \
            config.priors, False), kwargs={"model":"disk", "ncpus":ncpus, \
            "timelimit":args.timelimit, "ncpus_highmass":ncpus_highmass, \
            "with_hyperion":args.withhyperion, "source":source, "nice":nice, \
            "verbose":args.verbose, "ftcode":args.ftcode}, \
            pool=pool, backend=backend)
else:
    sampler = backend

# Run a few burner steps.

while sampler.iteration < config.max_nsteps:
    if args.action == "run":
        for sample in sampler.sample(sample, iterations=config.steps_per_iter):
            # Plot the current state of the chain.

            chain = sampler.get_chain()

            for j in range(ndim):
                fig, ax = plt.subplots(nrows=1, ncols=1)

                for k in range(config.nwalkers):
                    ax.plot(chain[:,k,j])

                plt.savefig("steps_{0:s}.png".format(keys[j]))

                plt.close(fig)

    # Get the best fit parameters and uncertainties from the last nplot steps.

    samples = sampler.get_chain(discard=sampler.iteration-config.nplot, \
            flat=True)

    # Make the cuts specified by the user.

    for command in trim:
        command = command.split(" ")

        for i, key in enumerate(keys):
            if key == command[0]:
                if command[1] == '<':
                    good = samples[:,i] > float(command[2])
                else:
                    good = samples[:,i] < float(command[2])

                samples = samples[good,:]

    # Get the best fit parameters.

    params = numpy.median(samples, axis=0)
    sigma = astropy.stats.mad_std(samples, axis=0)

    # Write out the results.

    f = open("fit.txt", "w")
    f.write("Best fit parameters:\n\n")
    for j in range(len(keys)):
        f.write("{0:s} = {1:f} +/- {2:f}\n".format(keys[j], params[j], \
                sigma[j]))
    f.write("\n")
    f.close()

    os.system("cat fit.txt")

    # Plot histograms of the resulting parameters.

    fig = corner.corner(samples, labels=labels, truths=params)

    plt.savefig("fit.pdf")

    plt.close(fig)

    # Make a dictionary of the best fit parameters.

    params = dict(zip(keys, params))

    ############################################################################
    #
    # Plot the results.
    #
    ############################################################################

    # Plot the best fit model over the data.

    fig, ax = plt.subplots(nrows=2*len(visibilities["file"]), ncols=3)

    # Create a high resolution model for averaging.

    m = modeling.run_disk_model(visibilities, images, spectra, params, \
            config.parameters, plot=True, ncpus=ncpus, \
            ncpus_highmass=ncpus_highmass, with_hyperion=args.withhyperion, \
            timelimit=args.timelimit, source=source, nice=nice, \
            ftcode=args.ftcode)

    # Plot the millimeter data/models.

    for j in range(len(visibilities["file"])):
        # Plot the visibilities.

        plotting.plot_1D_visibilities(visibilities, m, config.parameters, \
                params, index=j, fig=(fig, ax[2*j,0]))

        # Plot the 2D visibilities.

        plotting.plot_2D_visibilities(visibilities, m, config.parameters, \
                params, index=j, fig=(fig, ax[2*j+1,0:2]))

        # Plot the model image.

        plotting.plot_continuum_image(visibilities, m, config.parameters, \
                params, index=j, fig=(fig, ax[2*j,1]))

    # Plot the SED.

    plotting.plot_SED(spectra, m, SED=args.SED, fig=(fig, ax[0,2]))

    # Plot the scattered light image.

    for j in range(len(images["file"])):
        plotting.plot_scattered_light(visibilities, m, config.parameters, \
                params, index=j, fig=(fig, ax[1,2]))

    # Turn off axes when they aren't being used.

    if len(images["file"]) == 0:
        ax[1,2].set_axis_off()

    for j in range(len(visibilities["file"])):
        if j > 0:
            ax[2*j,2].set_axis_off()
            ax[2*j+1,2].set_axis_off()

    # Adjust the plot.

    fig.set_size_inches((12.5,8*len(visibilities["file"])))
    fig.subplots_adjust(left=0.07, right=0.98, top=0.95, bottom=0.08, \
            wspace=0.25, hspace=0.2)

    # Save the figure.

    fig.savefig("model.pdf")

    plt.close(fig)

    # If we're just plotting, make sure we don't loop forever.

    if args.action == "plot":
        break

# Now we can close the pool and end the code.

if args.action == "run":
    if withmpi:
        pool.close()



from scipy.optimize import minimize
from ..imaging import Image
from . import interpolate_model
import numpy

def rmlimage(data, imsize=512, pixelsize=0.01):

    def neg_ln_like(p):
        if numpy.any(p <= 0):
            return numpy.inf
        else:
            model = Image(numpy.array(p).reshape((imsize,imsize,1,1)), \
                    x=numpy.arange(imsize)*pixelsize, y=numpy.arange(imsize)*\
                    pixelsize, freq=data.freq)

            model_vis = interpolate_model(data.u, data.v, data.freq, model)

            return ((data.real - model_vis.real)**2 * data.weights + \
                    (data.imag - model_vis.imag)**2 * data.weights).sum() + \
                    (model.image * numpy.log(model.image)).sum()

    result = minimize(neg_ln_like, numpy.ones(imsize**2))

    return Image(numpy.array(result.x).reshape((imsize,imsize,1,1)), \
            x=numpy.arange(imsize)*pixelsize, y=numpy.arange(imsize)*\
            pixelsize, freq=data.freq)



from math import factorial

def fisher_exact(a,b,c,d):

    return float(factorial(a+b))*float(factorial(c+d))*float(factorial(a+c))* \
            float(factorial(b+d))/(float(factorial(a))*float(factorial(b))* \
            float(factorial(c))*float(factorial(d))*float(factorial(a+b+c+d)))



from .base_parameters import base_parameters

def check_parameters(parameters, nvis=3):
    # Make sure the code is backwards compatible to a time when only a single 
    # gas file was being supplied.

    if "gas_file" in parameters:
        parameters["gas_file1"] = parameters["gas_file"]
        parameters["logabundance1"] = parameters["logabundance"]

    # Make sure that the envelope dust is the same as the disk dust, if it is 
    # not specified.

    if "dust_file" in parameters and not "envelope_dust" in parameters:
        parameters["envelope_dust"] = parameters["dust_file"]

    # Now loop through and make sure all of the parameters are present and 
    # accounted for.

    for key in base_parameters:
        if not key in parameters:
            parameters[key] = base_parameters[key]

    # Make sure there are enough instances of freezeout for all of the gas 
    # files provided.

    index = 1
    while index > 0:
        if "gas_file"+str(index) in parameters:
            if not "logabundance{0:d}".format(index) in parameters:
                parameters["logabundance"+str(index)] = \
                        base_parameters["logabundance1"]
            if not "freezeout{0:d}".format(index) in parameters:
                parameters["freezeout"+str(index)] = \
                        base_parameters["freezeout1"]

            index += 1
        else:
            index = -1

    # Make sure there are enough instances of flux_unc for all of the 
    # data files that are being fit.

    for i in range(nvis):
        if not "flux_unc{0:d}".format(i+1) in parameters:
            parameters["flux_unc{0:d}".format(i+1)] = parameters["flux_unc1"]

    return parameters



import numpy
import h5py
from ..dust import Dust
from ..gas import Gas
from .Star import Star

class Grid:

    def __init__(self):
        self.density = []
        self.dust = []
        self.temperature = []
        self.stars = []
        self.number_density = []
        self.gas_temperature = []
        self.gas = []
        self.velocity = []
        self.microturbulence = []
        self.scattering_phase_freq = []
        self.scattering_phase = []

    def add_density(self, density, dust):
        self.density.append(density)
        self.dust.append(dust)

    def add_star(self, star):
        self.stars.append(star)

    def add_temperature(self, temperature):
        self.temperature.append(temperature)

    def add_number_density(self, number_density, gas):
        self.number_density.append(number_density)
        self.gas.append(gas)

    def add_gas_temperature(self, gas_temperature):
        self.gas_temperature.append(gas_temperature)

    def add_velocity(self, velocity):
        self.velocity.append(velocity)

    def add_microturbulence(self, microturbulence):
        self.microturbulence.append(microturbulence)

    def add_scattering_phase(self, freq, scattering_phase):
        self.scattering_phase_freq = freq
        self.scattering_phase = scattering_phase

    def set_cartesian_grid(self, w1, w2, w3):
        self.coordsystem = "cartesian"

        self.x = 0.5*(w1[0:w1.size-1] + w1[1:w1.size])
        self.y = 0.5*(w2[0:w2.size-1] + w2[1:w2.size])
        self.z = 0.5*(w3[0:w3.size-1] + w3[1:w3.size])

        self.w1 = w1
        self.w2 = w2
        self.w3 = w3

    def set_cylindrical_grid(self, w1, w2, w3):
        self.coordsystem = "cylindrical"

        self.rho = 0.5*(w1[0:w1.size-1] + w1[1:w1.size])
        self.phi = 0.5*(w2[0:w2.size-1] + w2[1:w2.size])
        self.z = 0.5*(w3[0:w3.size-1] + w3[1:w3.size])

        self.w1 = w1
        self.w2 = w2
        self.w3 = w3

    def set_spherical_grid(self, w1, w2, w3):
        self.coordsystem = "spherical"

        self.r = 0.5*(w1[0:w1.size-1] + w1[1:w1.size])
        self.theta = 0.5*(w2[0:w2.size-1] + w2[1:w2.size])
        self.phi = 0.5*(w3[0:w3.size-1] + w3[1:w3.size])

        self.w1 = w1
        self.w2 = w2
        self.w3 = w3

    def set_wavelength_grid(self, lmin, lmax, nlam, log=False):
        if log:
            self.lam = numpy.logspace(numpy.log10(lmin), numpy.log10(lmax), \
                    nlam)
        else:
            self.lam = numpy.linspace(lmin, lmax, nlam)

    def read(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        coordsystem = f['coordsystem'][()]
        w1 = f['w1'][...]
        w2 = f['w2'][...]
        w3 = f['w3'][...]

        if type(coordsystem) == bytes:
            coordsystem = coordsystem.decode('utf-8')

        if (coordsystem == 'cartesian'):
            self.set_cartesian_grid(w1, w2, w3)
        elif (coordsystem == 'cylindrical'):
            self.set_cylindrical_grid(w1, w2, w3)
        elif (coordsystem == 'spherical'):
            self.set_spherical_grid(w1, w2, w3)

        density = f['Density']
        for name in density:
            self.density.append(density[name][...])

        dust = f['Dust']
        for name in dust:
            d = Dust()
            d.set_properties_from_file(usefile=dust[name])
            self.dust.append(d)

        temperature = f['Temperature']
        for name in temperature:
            self.temperature.append(temperature[name][...])

        stars = f['Stars']
        for name in stars:
            star = Star()
            star.read(usefile=stars[name])
            self.stars.append(star)

        number_density = f['NumberDensity']
        for name in number_density:
            self.number_density.append(number_density[name][...])

        if 'GasTemperature' in f:
            gas_temperature = f['GasTemperature']
            for name in gas_temperature:
                self.gas_temperature.append(gas_temperature[name][...])

        if 'Microturbulence' in f:
            microturbulence = f['Microturbulence']
            for name in microturbulence:
                self.microturbulence.append(microturbulence[name][...])

        if "ScatteringPhase" in f:
            scattering_phase = f["ScatteringPhase"]
            self.scattering_phase = scattering_phase["ScatteringPhase"][...]
            self.scattering_phase_freq = \
                    scattering_phase["ScatteringPhaseFreq"][...]

        gas = f['Gas']
        for name in gas:
            g = Gas()
            g.set_properties_from_file(usefile=gas[name])
            self.gas.append(g)

        velocity = f['Velocity']
        for name in velocity:
            self.velocity.append(velocity[name][...])

        if ('lam' in f):
            self.lam = f['lam'][...]

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        f['coordsystem'] = self.coordsystem
        w1_dset = f.create_dataset("w1", (self.w1.size,), dtype='f')
        w1_dset[...] = self.w1
        w2_dset = f.create_dataset("w2", (self.w2.size,), dtype='f')
        w2_dset[...] = self.w2
        w3_dset = f.create_dataset("w3", (self.w3.size,), dtype='f')
        w3_dset[...] = self.w3

        density = f.create_group("Density")
        density_dsets = []
        for i in range(len(self.density)):
            density_dsets.append(density.create_dataset( \
                    "Density{0:d}".format(i), self.density[i].shape, dtype='f'))
            density_dsets[i][...] = self.density[i]

        dust = f.create_group("Dust")
        dust_groups = []
        for i in range(len(self.dust)):
            dust_groups.append(dust.create_group("Dust{0:d}".format(i)))
            self.dust[i].write(usefile=dust_groups[i])

        temperature = f.create_group("Temperature")
        temperature_dsets = []
        for i in range(len(self.temperature)):
            temperature_dsets.append(temperature.create_dataset( \
                    "Temperature{0:d}".format(i), self.temperature[i].shape, \
                    dtype='f'))
            temperature_dsets[i][...] = self.temperature[i]

        stars = f.create_group("Stars")
        stars_groups = []
        for i in range(len(self.stars)):
            stars_groups.append(stars.create_group("Star{0:d}".format(i)))
            self.stars[i].write(usefile=stars_groups[i])

        number_density = f.create_group("NumberDensity")
        number_density_dsets = []
        for i in range(len(self.number_density)):
            number_density_dsets.append(number_density.create_dataset( \
                    "NumberDensity{0:d}".format(i), \
                    self.number_density[i].shape, dtype='f'))
            number_density_dsets[i][...] = self.number_density[i]

        gas_temperature = f.create_group("GasTemperature")
        gas_temperature_dsets = []
        for i in range(len(self.gas_temperature)):
            gas_temperature_dsets.append(gas_temperature.create_dataset( \
                    "GasTemperature{0:d}".format(i), \
                    self.gas_temperature[i].shape, dtype='f'))
            gas_temperature_dsets[i][...] = self.gas_temperature[i]

        microturbulence = f.create_group("Microturbulence")
        microturbulence_dsets = []
        for i in range(len(self.microturbulence)):
            microturbulence_dsets.append(microturbulence.create_dataset( \
                    "Microturbulence{0:d}".format(i), \
                    self.microturbulence[i].shape, dtype='f'))
            microturbulence_dsets[i][...] = self.microturbulence[i]

        if len(self.scattering_phase) > 0:
            scattering_phase = f.create_group("ScatteringPhase")
            scattering_phase_dsets = []
            scattering_phase_dsets.append(scattering_phase.create_dataset( \
                    "ScatteringPhase", self.scattering_phase.shape, dtype='f'))
            scattering_phase_dsets[0][...] = self.scattering_phase
            scattering_phase_dsets.append(scattering_phase.create_dataset( \
                    "ScatteringPhaseFreq", self.scattering_phase_freq.shape, \
                    dtype='f'))
            scattering_phase_dsets[1][...] = self.scattering_phase_freq

        gas = f.create_group("Gas")
        gas_groups = []
        for i in range(len(self.gas)):
            gas_groups.append(gas.create_group("Gas{0:d}".format(i)))
            self.gas[i].write(usefile=gas_groups[i])

        velocity = f.create_group("Velocity")
        velocity_dsets = []
        for i in range(len(self.velocity)):
            velocity_dsets.append(velocity.create_dataset("Velocity{0:d}". \
                    format(i), self.velocity[i].shape, dtype='f'))
            velocity_dsets[i][...] = self.velocity[i]

        if hasattr(self, 'lam'):
            lam_dset = f.create_dataset("lam", self.lam.shape, dtype='f')
            lam_dset[...] = self.lam

        if (usefile == None):
            f.close()



import numpy
import astropy
import h5py
from ..constants.physics import c

class Image:

    def __init__(self,image=None,x=None,y=None,header=None,wave=None, \
            freq=None,unc=None,velocity=None):

        if (type(image) != type(None)):
            self.image = image

        if (type(x) != type(None)):
            self.x = x
            self.y = y
        if (type(header) != type(None)):
            self.header = header
        if (type(unc) != type(None)):
            self.unc = unc
        if (type(velocity) != type(None)):
            self.velocity = velocity

        if (type(wave) == type(None)) and (type(freq) != type(None)):
            self.freq = freq
            self.wave = c / freq
        elif (type(wave) != type(None)) and (type(freq) == type(None)):
            self.wave = wave
            self.freq = c / wave
        elif (type(wave) != type(None)) and (type(freq) != type(None)):
            self.wave = wave
            self.freq = freq

    def asFITS(self):
        hdulist = astropy.io.fits.HDUList([])
        for i in range(self.image[0,0,:].size):
            hdu = astropy.io.fits.PrimaryHDU(self.image[:,:,i])

            if hasattr(self, "header"):
                hdu.header = self.header[i]

            hdulist.append(hdu)

        return hdulist

    def read(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        if ('x' in f):
            x = f['x'][...]
            y = f['y'][...]
        else:
            x = None
            y = None

        if ('freq' in f):
            freq = f['freq'][...]
        else:
            freq = None

        image = f['image'][...]

        if ('unc' in f):
            unc = f['unc'][...]
        else:
            unc = None

        self.__init__(image, x=x, y=y, unc=unc, freq=freq)

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        if hasattr(self, "x"):
            x_dset = f.create_dataset("x", (self.x.size,), dtype='f')
            x_dset[...] = self.x
            y_dset = f.create_dataset("y", (self.y.size,), dtype='f')
            y_dset[...] = self.y

        if hasattr(self, "freq"):
            freq_dset = f.create_dataset("freq", (self.freq.size,), dtype='f')
            freq_dset[...] = self.freq

        image_dset = f.create_dataset("image", self.image.shape, dtype='f')
        image_dset[...] = self.image

        if hasattr(self, "unc"):
            unc_dset = f.create_dataset("uncertainty", self.unc.shape, \
                    dtype='f')
            unc_dset[...] = self.unc

        if (usefile == None):
            f.close()



from .spectroscopy import Spectrum
from .find_lines import find_lines
from .line_flux import line_flux
from .read_spectrum import read_spectrum
from .btsettl_photometry import btsettl_photometry
from .freefree import freefree



import numpy
import scipy
import h5py
from ..constants.physics import c
from ..constants.math import pi

class PAH:

    def calculate_size_distribution_opacity(self, amin, amax, p):
        na = int(round(numpy.log10(amax) - numpy.log10(amin))*100+1)
        a = numpy.logspace(numpy.log10(amin),numpy.log10(amax),na)
        kabsgrid = numpy.zeros((self.lam.size,na))
        kscagrid = numpy.zeros((self.lam.size,na))
        
        normfunc = a**(3-p)

        for i in range(na):
            self.calculate_opacity(a[i])
            
            kabsgrid[:,i] = self.kabs*normfunc[i]
            kscagrid[:,i] = self.ksca*normfunc[i]
        
        norm = scipy.integrate.trapz(normfunc,x=a)
        
        self.kabs = scipy.integrate.trapz(kabsgrid,x=a)/norm
        self.ksca = scipy.integrate.trapz(kscagrid,x=a)/norm
        self.kext = self.kabs + self.ksca
        self.albedo = self.ksca / self.kext

    def calculate_opacity(self, a):
        mdust = 4*pi*a**3/3*self.rho

        fabs = scipy.interpolate.interp2d(self.lam, self.a, \
                numpy.log10(self.Qabs), kind='linear')
        fsca = scipy.interpolate.interp2d(self.lam, self.a, \
                numpy.log10(self.Qsca), kind='linear')
        fext = scipy.interpolate.interp2d(self.lam, self.a, \
                numpy.log10(self.Qext), kind='linear')

        Qabs = 10.**fabs(self.lam, a)
        Qsca = 10.**fsca(self.lam, a)
        Qext = 10.**fext(self.lam, a)
        
        self.kabs = pi*a**2*Qabs/mdust
        self.ksca = pi*a**2*Qsca/mdust
        self.kext = pi*a**2*Qext/mdust
    
        self.albedo = self.ksca / self.kext

    def set_properties_from_draine(self, neutral, ion, qion):
        # Read in the neutral PAH data.

        f = open(neutral, "r")
        lines = f.readlines()
        f.close()

        self.nr = int(lines[7].split()[0])
        self.nlam = int(lines[8].split()[0])

        a_neutral = []
        optical_data_neutral = []
        for i in range(self.nr):
            a_neutral += [float(lines[10 + i*(self.nlam+3)].split()[0])]
            optical_data_neutral += [numpy.array([line.split() for line in \
                    lines[12+i*(self.nlam+3):12+i*(self.nlam+3)+self.nlam]], \
                    dtype=float)]

        a_neutral = numpy.array(a_neutral)
        optical_data_neutral = numpy.array(optical_data_neutral)
        lam_neutral = optical_data_neutral[:,:,0].mean(axis=0)

        # Read in the ionized PAH data.

        f = open(ion, "r")
        lines = f.readlines()
        f.close()

        nr = int(lines[7].split()[0])
        nlam = int(lines[8].split()[0])
        if nr != self.nr or (nlam != self.nlam):
            raise Warning("Ionized and neutral PAH files don't have the same "
                    "shape")

        a_ion = []
        optical_data_ion = []
        for i in range(self.nr):
            a_ion += [float(lines[10 + i*(self.nlam+3)].split()[0])]
            optical_data_ion += [numpy.array([line.split() for line in \
                    lines[12+i*(self.nlam+3):12+i*(self.nlam+3)+self.nlam]], \
                    dtype=float)]

        a_ion = numpy.array(a_ion)
        optical_data_ion = numpy.array(optical_data_ion)
        lam_ion = optical_data_ion[:,:,0].mean(axis=0)

        # Check that a and lam are the same between the two files.

        if not numpy.all(numpy.abs(a_ion - a_neutral)/a_neutral < 1.0e-6):
            raise Warning("Grain sizes are different between ionized and "
                    "neutral PAH data files.")
        else:
            self.a = 0.5*(a_ion + a_neutral) / 1.0e4

        if not numpy.all(numpy.abs(lam_ion - lam_neutral)/lam_neutral < 1.0e-6):
            raise Warning("Wavelengths are different between ionized and "
                    "neutral PAH data files.")
        else:
            self.lam = 0.5*(lam_ion + lam_neutral)[::-1] / 1.0e4

        # Load the ionization fraction data.

        data = numpy.loadtxt(qion, delimiter=",")
        qion_a = data[:,0]/1.0e8
        qion_q = data[:,1]

        # Merge the ionized and neutral data together based on qion.

        fqion = scipy.interpolate.interp1d(numpy.log10(qion_a), qion_q, \
                kind="linear", bounds_error=False, fill_value="extrapolate")

        q = fqion(numpy.log10(self.a))[:,numpy.newaxis,numpy.newaxis]

        optical_data = optical_data_ion*q + optical_data_neutral*(1-q)

        # Store the results in this object.

        self.Qext = optical_data[:,::-1,1]
        self.Qabs = optical_data[:,::-1,2]
        self.Qsca = optical_data[:,::-1,3]
        self.g = optical_data[:,::-1,4]

    def set_density(self, rho):
        self.rho = rho

    def set_properties_from_file(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        self.a = f['a'][...]
        self.lam = f['lam'][...]
        self.nu = c / self.lam

        if ('Qabs' in f):
            self.Qabs = f['Qabs'][...]
        if ('Qsca' in f):
            self.Qsca = f['Qsca'][...]
        if ('Qext' in f):
            self.Qext = f['Qext'][...]
        if ('g' in f):
            self.g = f['g'][...]

        if ('rho' in f):
            self.rho = f['rho'][...][0]

        if ('kabs' in f):
            self.kabs = f['kabs'][...]
        if ('ksca' in f):
            self.ksca = f['ksca'][...]
        if (hasattr(self, 'kabs') and hasattr(self, 'ksca')):
            self.kext = self.kabs + self.ksca
            self.albedo = self.ksca / self.kext

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        if hasattr(self, 'lam'):
            lam_dset = f.create_dataset("lam", (self.lam.size,), dtype='f')
            lam_dset[...] = self.lam
        if hasattr(self, 'a'):
            a_dset = f.create_dataset("a", (self.a.size,), dtype='f')
            a_dset[...] = self.a

        if hasattr(self, 'Qabs'):
            Qabs_dset = f.create_dataset("Qabs", self.Qabs.shape, dtype='f')
            Qabs_dset[...] = self.Qabs
        if hasattr(self, 'Qsca'):
            Qsca_dset = f.create_dataset("Qsca", self.Qsca.shape, dtype='f')
            Qsca_dset[...] = self.Qsca
        if hasattr(self, 'Qext'):
            Qext_dset = f.create_dataset("Qext", self.Qext.shape, dtype='f')
            Qext_dset[...] = self.Qext
        if hasattr(self, 'g'):
            g_dset = f.create_dataset("g", self.g.shape, dtype='f')
            g_dset[...] = self.g
        
        if hasattr(self, 'rho'):
            rho_dset = f.create_dataset("rho", (1,), dtype='f')
            rho_dset[...] = [self.rho]

        if hasattr(self, 'kabs'):
            kabs_dset = f.create_dataset("kabs", (self.kabs.size,), dtype='f')
            kabs_dset[...] = self.kabs
        if hasattr(self, 'ksca'):
            ksca_dset = f.create_dataset("ksca", (self.ksca.size,), dtype='f')
            ksca_dset[...] = self.ksca

        if (usefile == None):
            f.close()



import numpy
import h5py
from ..constants.physics import sigma
from ..constants.astronomy import L_sun, R_sun

class Star:
    def __init__(self, mass=0.5, luminosity=1.0, temperature=4000., x=0.0, \
            y=0.0, z=0.0):
        self.mass = mass
        self.luminosity = luminosity
        self.temperature = temperature
        self.radius = (luminosity*L_sun/ \
                (4*numpy.pi*sigma*temperature**4))**(1./2)/R_sun
        self.x = x
        self.y = y
        self.z = z

    def read(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        self.mass = f['mass'][()]
        self.luminosity = f['luminosity'][()]
        self.temperature = f['temperature'][()]
        self.radius = f['radius'][()]
        self.x = f['x'][()]
        self.y = f['y'][()]
        self.z = f['z'][()]

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        f['mass'] = self.mass
        f['luminosity'] = self.luminosity
        f['temperature'] = self.temperature
        f['radius'] = self.radius

        f['x'] = self.x
        f['y'] = self.y
        f['z'] = self.z

        if (usefile == None):
            f.close()



import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy

def plot_scattered_light(images, model, parameters, params, index=0, \
        fig=None):

    # If no axes are provided, create them.

    if fig == None:
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(4.5,4))
    else:
        fig, ax = fig

    # Scale the image appropriately.

    c = scale_image(images["data"][index], mode="arcsinh")

    # Show the image.

    ax.imshow(c[:,:,0,0], origin="lower", interpolation="nearest", cmap="gray")

    # Scale the model image.

    c = scale_image(model.images[images["lam"][index]], \
            mode=images["plot_mode"][index])

    # Contour the model image over the data.

    levels = numpy.array([0.05,0.25,0.45,0.65,0.85,1.0]) * \
            (c.max() - c.min()) + c.min()

    ax.contour(c[:,:,0,0], colors='gray', levels=levels)

    # Adjust the ticks.

    transform3 = ticker.FuncFormatter(Transform(0, images["npix"][index], \
            images["pixelsize"][index], '%.1f"'))

    ticks = images["ticks"][index]

    ax.set_xticks(images["npix"][index]/2+ticks/images["pixelsize"][index])
    ax.set_yticks(images["npix"][index]/2+ticks/images["pixelsize"][index])
    ax.get_xaxis().set_major_formatter(transform3)
    ax.get_yaxis().set_major_formatter(transform3)

    # Adjust the axes labels.

    ax.set_xlabel("$\Delta$RA")
    ax.set_ylabel("$\Delta$Dec")

    # Return the figure and axes.

    return fig, ax

# Define a useful class for plotting.

class Transform:
    def __init__(self, xmin, xmax, dx, fmt):
        self.xmin = xmin
        self.xmax = xmax
        self.dx = dx
        self.fmt = fmt

    def __call__(self, x, p):
        return self.fmt% ((x-(self.xmax-self.xmin)/2)*self.dx)

# Define a function to scale an image to look nice.

def scale_image(image, mode="linear"):
    vmin = image.image.min()
    vmax = numpy.percentile(image.image, 95)

    a = 1000.
    b = (image.image - vmin) / (vmax - vmin)

    if mode == "linear":
        c = b
    elif mode == "arcsinh":
        c = numpy.arcsinh(10*b)/3.
    elif mode == "log":
        c = numpy.log10(a*b+1)/numpy.log10(a)
    else:
        print("Not a valid mode!")

    return c



from . import __file__
from scipy.interpolate import interp1d
import numpy
import os

def redden(wave, flux, Av, law="steenman", Rv='3.1', magnitudes=False):

    # Read in the extinction coefficients.

    if law == 'steenman':
        data = numpy.loadtxt(os.path.dirname(__file__)+ \
                "/reddening/steenman_the.dat", usecols=[1,2])
    elif law == 'mcclure':
        if Av <= 1.:
            data = numpy.loadtxt(os.path.dirname(__file__)+ \
                    "/reddening/mcclure.dat", usecols=[0,1])
        else:
            data = numpy.loadtxt(os.path.dirname(__file__)+ \
                    "/reddening/mcclure.dat", usecols=[0,2])
    elif law == 'draine':
        if Rv == '3.1':
            data = numpy.loadtxt(os.path.dirname(__file__)+ \
                    "/reddening/draine_3.1.dat", skiprows=80, \
                    usecols=(0,3))

            data = data[::-1,:]
            data[:,1] /= data[605,1]

        elif Rv == '4.0':
            data = numpy.loadtxt(os.path.dirname(__file__)+ \
                    "/reddening/draine_4.0.dat", skiprows=80, \
                    usecols=(0,3))

            data = data[::-1,:]
            data[:,1] /= data[605,1]

        elif Rv == '5.1':
            data = numpy.loadtxt(os.path.dirname(__file__)+ \
                    "/reddening/draine_5.1.dat", skiprows=80, \
                    usecols=(0,3))

            data = data[::-1,:]
            data[:,1] /= data[605,1]

    waves = data[:,0]
    coeffs = data[:,1]

    # Interpolate the data to the wavelength range supplied.

    coeffs = interp1d(numpy.log10(waves), numpy.log10(coeffs), \
            fill_value="extrapolate")

    if magnitudes:
        return flux + Av * 10.**coeffs(numpy.log10(wave))
    else:
        coeff = 10.**coeffs(numpy.log10(wave))

        # Now adjust the provided flux.

        return flux / 10.**(Av * coeff / 2.5)



import numpy
import h5py
from scipy.integrate import trapz
from ..constants.physics import G, m_p
from ..constants.astronomy import AU, M_sun
from ..constants.math import pi
from ..dust import Dust
from ..gas import Gas

class Disk:
    def __init__(self, mass=1.0e-3, rmin=0.1, rmax=300, plrho=2.37, h0=0.1, \
            plh=58./45., t0=None, plt=None, dust=None, gap_rin=[], gap_rout=[],\
            gap_delta=[], tmid0=None, tatm0=None, zq0=None, pltgas=None, \
            delta=None, aturb=None, gaussian_gaps=False, gamma_taper=None):
        self.mass = mass
        self.rmin = rmin
        self.rmax = rmax
        self.plrho = plrho
        self.gamma_taper = gamma_taper
        self.h0 = h0
        self.plh = plh
        self.t0 = t0
        self.plt = plt
        self.gap_rin = gap_rin
        self.gap_rout = gap_rout
        self.gap_delta = gap_delta
        self.gaussian_gaps = gaussian_gaps
        if (dust != None):
            self.dust = dust

        # The gas parameter lists.

        self.gas = []
        self.abundance = []
        self.freezeout = []
        self.tmid0 = tmid0
        self.tatm0 = tatm0
        self.zq0 = zq0
        self.pltgas = pltgas
        self.delta = delta
        self.aturb = aturb

    def add_gas(self, gas, abundance, freezeout=0.):
        self.gas.append(gas)
        self.abundance.append(abundance)
        self.freezeout.append(freezeout)

    def density(self, x1, x2, x3, coordsys="spherical", normalize=True):
        ##### Set up the coordinates

        if coordsys == "spherical":
            rt, tt, pp = numpy.meshgrid(x1*AU, x2, x3, indexing='ij')

            rr = rt*numpy.sin(tt)
            zz = rt*numpy.cos(tt)
        elif coordsys == "cartesian":
            xx, yy, zz = numpy.meshgrid(x1*AU, x2*AU, x3*AU, indexing='ij')

            rr = (xx**2 + yy**2)**0.5
        elif coordsys == "cylindrical":
            rr, pp, zz = numpy.meshgrid(x1*AU, x2, x3*AU, indexing='ij')

        ##### Make the dust density model for a protoplanetary disk.

        Sigma = self.surface_density(rr/AU, normalize=normalize)
        h = self.scale_height(rr/AU)
        
        rho = Sigma / (numpy.sqrt(2*numpy.pi)*h) * numpy.exp(-0.5*(zz / h)**2)

        return rho

    def number_density(self, r, theta, phi, gas=0):
        rho = self.density(r, theta, phi)

        rho_gas = rho * 100

        rho_gas_critical = (100. / 0.8) * 2.37*m_p
        rho_gas[rho_gas < rho_gas_critical] = 1.0e-50

        n_H2 = rho_gas * 0.8 / (2.37*m_p)

        n = n_H2 * self.abundance[gas]

        return n

    def surface_density(self, r, normalize=True):
        # Get the disk parameters.

        rin = self.rmin * AU
        rout = self.rmax * AU
        mass = self.mass * M_sun
        gamma = self.plrho - self.plh

        # Set up the surface density.

        Sigma0 = (2-gamma)*mass/(2*pi*(1*AU)**(gamma)) / \
                (rout**(-gamma+2) - rin**(-gamma+2))

        Sigma = Sigma0 * r**(-gamma)

        Sigma[(r >= rout/AU) ^ (r <= rin/AU)] = 0e0

        # In case of r == 0 (a singularity), get the value from slightly off 0.

        dr = r[r > 0].min()
        Sigma[r == 0] = Sigma0 * (0.7*dr)**(-gamma)

        # Add gaps to the disk.

        for i in range(len(self.gap_rin)):
            if self.gaussian_gaps:
                gap_r = (self.gap_rin[i] + self.gap_rout[i])/2
                gap_w = self.gap_rout[i] - self.gap_rin[i]

                Sigma /= 1 + 1./self.gap_delta[i] * numpy.exp(-4*numpy.log(2.)*\
                        (r - gap_r)**2 / gap_w**2)
            else:
                Sigma[(r >= self.gap_rin[i]) & \
                        (r <= self.gap_rout[i])] *= self.gap_delta[i]
        
        ##### Normalize the surface density correctly.
        
        if normalize:
            r_high = numpy.logspace(numpy.log10(self.rmin), \
                    numpy.log10(self.rmax), 1000)
            Sigma_high = self.surface_density(r_high, normalize=False)

            scale = mass / (2*numpy.pi*trapz(r_high*AU*Sigma_high, r_high*AU))

            Sigma *= scale

        return Sigma

    def scale_height(self, r):
        return self.h0 * AU * r**self.plh

    def temperature(self, r, theta, phi):
        ##### Disk Parameters
        
        rin = self.rmin * AU
        rout = self.rmax * AU
        t0 = self.t0
        plt = self.plt

        ##### Set up the coordinates

        rt, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        rr = rt*numpy.sin(tt)
        zz = rt*numpy.cos(tt)

        ##### Make the dust density model for a protoplanetary disk.
        
        t = t0 * (rr / (1*AU))**(-plt)

        t[(rr >= rout) ^ (rr <= rin)] = 0e0

        t[t > 10000.] = 10000.
        
        return t

    def temperature_1d(self, r):
        rin = self.rmin * AU
        rout = self.rmax * AU
        t0 = self.t0
        plt = self.plt

        T = t0 * r**(-plt)

        T[(r >= rout/AU) ^ (r <= rin/AU)] = 0.0

        dr = r[r > 0].min()
        T[r == 0] = t0 * (0.7*dr)**(-plt)

        return T

    def microturbulence(self, r, theta, phi):
        ##### Disk Parameters
        
        rin = self.rmin * AU
        rout = self.rmax * AU
        t0 = self.t0
        plt = self.plt

        ##### Set up the coordinates

        rt, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        rr = rt*numpy.sin(tt)
        zz = rt*numpy.cos(tt)

        ##### Make the dust density model for a protoplanetary disk.
        
        aturb = numpy.ones(rr.shape)*self.aturb*1.0e5
        
        return aturb

    def velocity(self, r, theta, phi, mstar=0.5):
        mstar *= M_sun

        rt, tt, pp = numpy.meshgrid(r*AU, theta, phi,indexing='ij')

        rr = rt*numpy.sin(tt)
        zz = rt*numpy.cos(tt)

        v_r = numpy.zeros(rr.shape)
        v_theta = numpy.zeros(rr.shape)
        v_phi = numpy.sqrt(G*mstar*rr**2/rt**3)

        return numpy.array((v_r, v_theta, v_phi))

    def read(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        self.mass = f['mass'][()]
        self.rmin = f['rmin'][()]
        self.rmax = f['rmax'][()]
        self.plrho = f['plrho'][()]
        self.h0 = f['h0'][()]
        self.plh = f['plh'][()]
        if 'gamma_taper' in f:
            self.gamma_taper = f['gamma_taper']

        if 't0' in f:
            self.t0 = f['t0'][()]
            self.plt = f['plt'][()]

        if 'tmid0' in f:
            self.tmid0 = f['tmid0'][()]
            self.tatm0 = f['tatm0'][()]
            self.zq0 = f['zq0'][()]
            self.pltgas = f['pltgas'][()]
            self.delta = f['delta'][()]

        if 'aturb' in f:
            self.aturb = f['aturb'][()]

        if ('Dust' in f):
            self.dust = Dust()
            self.dust.set_properties_from_file(usefile=f['Dust'])

        if ('Gas' in f):
            for name in f['Gas']:
                self.gas.append(Gas())
                self.abundance.append(f['Gas'][name]['Abundance'][()])
                self.gas[-1].set_properties_from_file(usefile=f['Gas'][name])

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        f['mass'] = self.mass
        f['rmin'] = self.rmin
        f['rmax'] = self.rmax
        f['plrho'] = self.plrho
        f['h0'] = self.h0
        f['plh'] = self.plh

        if self.gamma_taper != None:
            f['gamma_taper'] = self.gamma_taper

        if self.t0 != None:
            f['t0'] = self.t0
            f['plt'] = self.plt

        if self.tmid0 != None:
            f['tmid0'] = self.tmid0
            f['tatm0'] = self.tatm0
            f['zq0'] = self.zq0
            f['pltgas'] = self.pltgas
            f['delta'] = self.delta

        if self.aturb != None:
            f['aturb'] = self.aturb

        if hasattr(self, 'dust'):
            dust = f.create_group("Dust")
            self.dust.write(usefile=dust)

        gases = []
        if hasattr(self, 'gas'):
            gas = f.create_group("Gas")
            for i in range(len(self.gas)):
                gases.append(gas.create_group("Gas{0:d}".format(i)))
                gases[i]["Abundance"] = self.abundance[i]
                gases[i]["Freezeout"] = self.freezeout[i]
                self.gas[i].write(usefile=gases[i])

        if (usefile == None):
            f.close()



import numpy
import scipy
import h5py
from ..constants.physics import c
from ..constants.math import pi
from .bhmie import bhmie
from .bhcoat import bhcoat
from .dmilay import dmilay

class Dust:

    def add_coat(self, coat):
        self.coat = coat

    def calculate_optical_constants_on_wavelength_grid(self, lam):
        f = scipy.interpolate.interp1d(self.lam, self.n)
        n = f(lam)
        f = scipy.interpolate.interp1d(self.lam, self.k)
        k = f(lam)

        self.lam = lam
        self.nu = c / self.lam

        self.n = n
        self.k = k
        self.m = self.n + 1j*self.k

    def calculate_opacity_at_wavelength(self, lam):
        f = scipy.interpolate.interp1d(self.lam, self.kabs)
        kabs_interp = f(lam)

        return kabs_interp

    def calculate_size_distribution_opacity(self, amin, amax, p, \
            coat_volume_fraction=0.0, nang=1000, with_dhs=False, fmax=0.8, \
            nf=50):
        na = int(round(numpy.log10(amax) - numpy.log10(amin))*100+1)
        a = numpy.logspace(numpy.log10(amin),numpy.log10(amax),na)
        kabsgrid = numpy.zeros((self.lam.size,na))
        kscagrid = numpy.zeros((self.lam.size,na))
        
        normfunc = a**(3-p)

        for i in range(na):
            if with_dhs:
                self.calculate_dhs_opacity(a[i], fmax=fmax, nf=nf, nang=nang)
            else:
                self.calculate_opacity(a[i], \
                        coat_volume_fraction=coat_volume_fraction, nang=nang)
            
            kabsgrid[:,i] = self.kabs*normfunc[i]
            kscagrid[:,i] = self.ksca*normfunc[i]
        
        norm = scipy.integrate.trapz(normfunc,x=a)
        
        self.kabs = scipy.integrate.trapz(kabsgrid,x=a)/norm
        self.ksca = scipy.integrate.trapz(kscagrid,x=a)/norm
        self.kext = self.kabs + self.ksca
        self.albedo = self.ksca / self.kext

    def calculate_opacity(self, a, coat_volume_fraction=0.0, nang=1000):
        self.kabs = numpy.zeros(self.lam.size)
        self.ksca = numpy.zeros(self.lam.size)
        
        if not hasattr(self, 'coat'):
            mdust = 4*pi*a**3/3*self.rho
            
            for i in range(self.lam.size):
                x = 2*pi*a/self.lam[i]
                
                S1,S2,Qext,Qsca,Qback,gsca=bhmie(x,self.m[i],nang)
                
                Qabs = Qext - Qsca
                
                self.kabs[i] = pi*a**2*Qabs/mdust
                self.ksca[i] = pi*a**2*Qsca/mdust
        
        else:
            a_coat = a*(1+coat_volume_fraction)**(1./3)

            mdust = 4*pi*a**3/3*self.rho+ \
                    4*pi/3*(a_coat**3-a**3)*self.coat.rho
            
            for i in range(self.lam.size):
                x = 2*pi*a/self.lam[i]
                y = 2*pi*a_coat/self.lam[i]
                
                Qext,Qsca,Qback=bhcoat(x,y,self.m[i],self.coat.m[i])
                
                Qabs = Qext - Qsca
                
                self.kabs[i] = pi*a_coat**2*Qabs/mdust
                self.ksca[i] = pi*a_coat**2*Qsca/mdust

        self.kext = self.kabs + self.ksca
        self.albedo = self.ksca / self.kext

    def calculate_dhs_opacity(self, a, fmax=0.8, nf=50, nang=1000):

        self.kabs = numpy.zeros(self.lam.size)
        self.ksca = numpy.zeros(self.lam.size)
        
        for i in range(self.lam.size):
            for j, f in enumerate(numpy.linspace(0., fmax, nf)):
                x = 2*pi*a*f**(1./3)/self.lam[i]
                y = 2*pi*a/self.lam[i]

                if f == 0:
                    S1,S2,Qext,Qsca,Qback,gsca=bhmie(y,self.m[i],nang)
                else:
                    Qext, Qsca, Qback, g, M1, M2, S21, D21 = dmilay(\
                            a*f**(1./3), a, 2*pi/self.lam[i], self.m[i].real - \
                            1j*self.m[i].imag, 1.0+1j*0.0, [0.], 1., 1)
                
                Qabs = Qext - Qsca
                
                mdust = 4*pi*a**3*(1.-f)/3*self.rho

                self.kabs[i] += pi*a**2*Qabs/mdust * 1./fmax * fmax/nf
                self.ksca[i] += pi*a**2*Qsca/mdust * 1./fmax * fmax/nf
    
        self.kext = self.kabs + self.ksca
        self.albedo = self.ksca / self.kext

    def set_density(self, rho):
        self.rho = rho

    def set_optical_constants(self, lam, n, k):
        self.lam = lam
        self.nu = c / self.lam

        self.n = n
        self.k = k
        self.m = n+1j*k

    def set_optical_constants_from_draine(self, filename):
        opt_data = numpy.loadtxt(filename)

        self.lam = numpy.flipud(opt_data[:,0])*1.0e-4
        self.nu = c / self.lam

        self.n = numpy.flipud(opt_data[:,3])+1.0
        self.k = numpy.flipud(opt_data[:,4])
        self.m = self.n+1j*self.k

    def set_optical_constants_from_henn(self, filename):
        opt_data = numpy.loadtxt(filename)

        self.lam = opt_data[:,0]*1.0e-4
        self.nu = c / self.lam

        self.n = opt_data[:,1]
        self.k = opt_data[:,2]
        self.m = self.n+1j*self.k

    def set_optical_constants_from_jena(self, filename, type="standard"):
        opt_data = numpy.loadtxt(filename)

        if type == "standard":
            self.lam = numpy.flipud(1./opt_data[:,0])
            self.n = numpy.flipud(opt_data[:,1])
            self.k = numpy.flipud(opt_data[:,2])
        elif type == "umwave":
            self.lam = numpy.flipud(opt_data[:,0])*1.0e-4
            self.n = numpy.flipud(opt_data[:,1])
            self.k = numpy.flipud(opt_data[:,2])

        self.nu = c / self.lam
        self.m = self.n+1j*self.k

    def set_optical_constants_from_oss(self, filename):
        opt_data = numpy.loadtxt(filename)
        
        self.lam = opt_data[:,0] # in cm
        self.nu = c / self.lam

        self.n = opt_data[:,1]
        self.k = opt_data[:,2]
        self.m = self.n+1j*self.k

    def set_properties(self, lam, kabs, ksca):
        self.lam = lam
        self.nu = c / self.lam

        self.kabs = kabs
        self.ksca = ksca
        self.kext = self.kabs + self.ksca
        self.albedo = self.ksca / self.kext

    def set_properties_from_file(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        self.lam = f['lam'][...]
        self.nu = c / self.lam

        if ('n' in f):
            self.n = f['n'][...]
        if ('k' in f):
            self.k = f['k'][...]
        if (hasattr(self, 'n') and hasattr(self, 'k')):
            self.m = self.n + 1j*self.k

        if ('rho' in f):
            self.rho = f['rho'][...][0]

        if ('kabs' in f):
            self.kabs = f['kabs'][...]
        if ('ksca' in f):
            self.ksca = f['ksca'][...]
        if (hasattr(self, 'kabs') and hasattr(self, 'ksca')):
            self.kext = self.kabs + self.ksca
            self.albedo = self.ksca / self.kext

        if ('coat' in f):
            self.coat = Dust()
            self.coat.set_properties_from_file(usefile=f['coat'])

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        if hasattr(self, 'lam'):
            lam_dset = f.create_dataset("lam", (self.lam.size,), dtype='f')
            lam_dset[...] = self.lam
        if hasattr(self, 'n'):
            n_dset = f.create_dataset("n", (self.n.size,), dtype='f')
            n_dset[...] = self.n
        if hasattr(self, 'k'):
            k_dset = f.create_dataset("k", (self.k.size,), dtype='f')
            k_dset[...] = self.k
        
        if hasattr(self, 'rho'):
            rho_dset = f.create_dataset("rho", (1,), dtype='f')
            rho_dset[...] = [self.rho]

        if hasattr(self, 'kabs'):
            kabs_dset = f.create_dataset("kabs", (self.kabs.size,), dtype='f')
            kabs_dset[...] = self.kabs
        if hasattr(self, 'ksca'):
            ksca_dset = f.create_dataset("ksca", (self.ksca.size,), dtype='f')
            ksca_dset[...] = self.ksca
        if hasattr(self, 'g'):
            g_dset = f.create_dataset("g", (self.g.size,), dtype='f')
            g_dset[...] = self.g

        if hasattr(self, 'coat'):
            coat_group = f.create_group("coat")

            self.coat.write(usefile=coat_group)

        if (usefile == None):
            f.close()



from numpy import zeros
from numpy.random import randint, normal

def change_params(params, sigma_params, change_param=None):

    new_params = zeros(params.size)

    if change_param == None:
        change_param = randint(low=0, high=params.size, size=1)[0]

    new_params = params.copy()
    new_params[change_param] = params[change_param] + \
        normal(loc=0., scale=sigma_params[change_param], size=1)[0]

    return new_params



from numpy import loadtxt
from .spectroscopy import Spectrum

def read_spectrum(file):
    
    spectrum = loadtxt(file)
    wave = spectrum[:,0]
    flux = spectrum[:,1]
    unc = spectrum[:,2]
    
    data = Spectrum(wave,flux,unc)
    
    return data



from ..constants.astronomy import arcsec
from ..modeling import check_parameters
from ..modeling import base_parameters
import importlib
import sys

def load_config(path=''):
    r"""
    Load a pdspy configuration file in as a module.

    Args:
        :attr:`path` (`str`):
            The path to the directory where the information for the radiative transfer model is stored.

    Returns:
        :attr:`config` (`module`):
            A python module containing the configuration information for pdspy modeling.
    """

    # Import the configuration file information.

    sys.path.append(path)

    import config
    importlib.reload(config)

    sys.path.remove(path)

    # If some set of information isn't supplied, provide it.

    if not hasattr(config, "visibilities"):
        config.visibilities = {\
            "file":[], \
            "pixelsize":[],\
            "freq":[],\
            "lam":[],\
            "npix":[],\
            "weight":[],\
            "x0":[],\
            "y0":[],\
            # Info for the image.
            "image_file":[],\
            "image_pixelsize":[],\
            "image_npix":[],\
            # Info for the plots.
            "nrows":[],\
            "ncols":[],\
            "ind0":[],\
            "ticks":[],\
            "image_ticks":[],\
            "fmt":[],\
            }

    if not hasattr(config, "images"):
        config.images = images = {\
            "file":[],\
            "npix":[],\
            "pixelsize":[],\
            "lam":[],\
            "bmaj":[],\
            "bmin":[],\
            "bpa":[],\
            "ticks":[],\
            "plot_mode":[],\
            }

    if not hasattr(config, "spectra"):
        config.spectra = {\
            "file":[],\
            "bin?":[],\
            "nbins":[],\
            "weight":[],\
            }

    if not hasattr(config, "parameters"):
        config.parameters = base_parameters

    if not hasattr(config, "priors"):
        config.priors = {}

    # Get the correct binsize and number of bins for averaging the visibility
    # data.

    config.visibilities["gridsize"] = []
    config.visibilities["binsize"] = []

    for i in range(len(config.visibilities["file"])):
        config.visibilities["gridsize"].append(config.visibilities["npix"][i])
        config.visibilities["binsize"].append(1/ \
                (config.visibilities["npix"][i]* \
                config.visibilities["pixelsize"][i]*arcsec))

    # Make sure the fmt keyword is in the visibilities dictionary.

    if not "fmt" in config.visibilities:
        config.visibilities["fmt"] = ['4.1f' for i in range(len(\
                config.visibilities["file"]))]

    # Make sure the visibilities dictionary has SPW, tolerance, and whether
    # corrected or data column in it.

    if not "spw" in config.visibilities:
        config.visibilities["spw"] = [[0] for i in range(len(\
                config.visibilities["file"]))]
    if not "tolerance" in config.visibilities:
        config.visibilities["tolerance"] = [0.01 for i in range(len(\
                config.visibilities["file"]))]
    if not "datacolumn" in config.visibilities:
        config.visibilities["datacolumn"] = ["corrected" for i in range(len(\
                config.visibilities["file"]))]

    # Make sure the visibilities dictionary has subsampling and averaging
    # parameters.

    if not "subsample" in config.visibilities:
        config.visibilities["subsample"] = [1 for i in range(len(\
                config.visibilities["file"]))]
    if not "averaging" in config.visibilities:
        config.visibilities["averaging"] = [1 for i in range(len(\
                config.visibilities["file"]))]
    if not "hanning" in config.visibilities:
        config.visibilities["hanning"] = [False for i in range(len(\
                config.visibilities["file"]))]

    # Make sure the spectra dictionary has a "weight" entry.

    if not "weight" in config.spectra:
        config.spectra["weight"] = [1. for i in range(len(\
                config.spectra["file"]))]

    # Check that all of the parameters are correct.

    config.parameters = check_parameters(config.parameters, \
            nvis=len(config.visibilities["file"]))

    # Return all of the configuration information.

    return config



from pdspy.constants.astronomy import Jy
from pdspy.constants.physics import c as c_l
import matplotlib.pyplot as plt

def plot_SED(spectra, model, SED=False, fig=None, model_color="g", 
        linewidth=1, fontsize="medium"):
    r"""
    Plot the SED generated by a radiative transfer modeling run with pdspy, typically generated by the output of `modeling.run_disk_model`.

    Args:
        :attr:`spectra` (list):
            List of `Spectrum` objects with data for the object you are studying.
        :attr:`model` (`modeling.Model`):
            The radiative transfer model that you would like to plot the SED of. The `Model.spectra` dictionary must include a `"SED"` key. Typically this is the output of modeling.run_disk_model.
        :attr:`SED` (bool, optional):
            Whether to plot as a traditional SED (`True`), i.e. as :math:`\nu F_{\nu}`, or as a spectrum, i.e. :math:`F_{\nu}`. Default: `False`
        :attr:`fig` (`tuple`, `(matplotlib.Figure, matplotlib.Axes)`, optional):
            If you've already created a figure and axes to put the plot in, you can supply them here. Otherwise, `plot_SED` will generate them for you. Default: `None`
        :attr:`model_color` (str, optional):
            The color to use for plotting the model SED. Default: `"g"`
        :attr:`linewidth` (int, optional):
            What linewidth to use for plotting the model. Default: 1
        :attr:`fontsize` (`str` or `int`):
            What fontsize to use for labels, ticks, etc. Default: `"medium"`

    Returns:
        :attr:`fig` (`matplotlib.Figure`):
            The matplotlib figure that was used for the plot.
        :attr:`ax` (`matplotlib.Axes`):
            The matplotlib axes that were used for the plot.
    """

    # If no axes are provided, create them.

    if fig == None:
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(4.5,4))
    else:
        fig, ax = fig

    # Plot the SED.

    for j in range(len(spectra["file"])):
        if spectra["bin?"][j]:
            if SED:
                ax.plot(spectra["data"][j].wave, \
                        c_l / spectra["data"][j].wave / 1.0e-4 * \
                        spectra["data"][j].flux * Jy, "k-")
            else:
                ax.plot(spectra["data"][j].wave, spectra["data"][j].flux, \
                        "k-")
        else:
            if SED:
                ax.errorbar(spectra["data"][j].wave, \
                        c_l / spectra["data"][j].wave / 1.0e-4 * \
                        spectra["data"][j].flux * Jy, fmt="ko", \
                        yerr=c_l / spectra["data"][j].wave / 1.0e-4 * \
                        spectra["data"][j].unc * Jy, markeredgecolor="k")
            else:
                ax.errorbar(spectra["data"][j].wave, \
                        spectra["data"][j].flux, fmt="ko", \
                        yerr=spectra["data"][j].unc, markeredgecolor="k")

    if len(spectra["file"]) > 0:
        if SED:
            ax.plot(model.spectra["SED"].wave, c_l/model.spectra["SED"].wave / \
                    1.0e-4 * model.spectra["SED"].flux * Jy, "-", \
                    color=model_color, linewidth=linewidth)
        else:
            ax.plot(model.spectra["SED"].wave, model.spectra["SED"].flux, "-", \
                    color=model_color, linewidth=linewidth)

    # Add axes labels and adjust the plot.

    if SED:
        ax.axis([0.1,1.0e4,1e-13,1e-8])
    else:
        ax.axis([0.1,1.0e4,1e-6,1e3])

    ax.set_xscale("log", nonpositive='clip')
    ax.set_yscale("log", nonpositive='clip')

    ax.set_xlabel("$\lambda$ [$\mu$m]", fontsize=fontsize)
    if SED:
        ax.set_ylabel(r"$\nu F_{\nu}$ [ergs s$^{-1}$ cm$^{-2}$]", \
                fontsize=fontsize)
    else:
        ax.set_ylabel(r"$F_{\nu}$ [Jy]", fontsize=fontsize)

    # Return the figure and axes.

    return fig, ax



import numpy
import astropy
import h5py
from ..constants.physics import c

class Spectrum:

    def __init__(self, wave=None, flux=None, unc=None):
        if (type(wave) != type(None)):
            self.wave = wave
            self.freq = c / wave / 1.0e-4
            self.flux = flux
            if type(unc) == type(None):
                unc = numpy.zeros(wave.size)
                self.unc = unc
            else:
                self.unc = unc

    def asFITS(self):
        hdulist = astropy.io.fits.HDUList([])

        hdu = pyfits.PrimaryHDU(numpy.concatenate(( \
                self.wave.reshape((1,self.wave.size)), \
                self.flux.reshape((1,self.wave.size)), \
                self.unc.reshape((1,self.wave.size)))))
        hdulist.append(hdu)

        return hdulist

    def read(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "r")
        else:
            f = usefile

        wave = f['wave'][...]
        flux = f['flux'][...]
        unc = f['unc'][...]

        self.__init__(wave, flux, unc)

        if (usefile == None):
            f.close()

    def write(self, filename=None, usefile=None):
        if (usefile == None):
            f = h5py.File(filename, "w")
        else:
            f = usefile

        if hasattr(self, "wave"):
            wave_dset = f.create_dataset("wave", (self.wave.size,), dtype='f')
            wave_dset[...] = self.wave
            flux_dset = f.create_dataset("flux", (self.flux.size,), dtype='f')
            flux_dset[...] = self.flux
            unc_dset = f.create_dataset("unc", (self.unc.size,), dtype='f')
            unc_dset[...] = self.unc

        if (usefile == None):
            f.close()



from .pms_evolutionary_tracks import pms_get_mstar
from .pms_evolutionary_tracks import pms_get_age
from .pms_evolutionary_tracks import pms_get_teff
from .pms_evolutionary_tracks import pms_get_luminosity
from .pms_evolutionary_tracks import pms_get_radius
from .Teff_from_SpT import Teff_from_SpT



from .column import MaskedColumn
import astropy.table
import astropy.utils
import h5py
import os

class Table(astropy.table.Table):

    def __init__(self, *args, **kwargs):
        super(Table, self).__init__(*args, **kwargs)

        self.MaskedColumn = MaskedColumn
        if self._masked:
            self._column_class = MaskedColumn

    @staticmethod
    def read_hdf5(*args, **kwargs):
        try:
            self = Table(astropy.table.Table.read(*args, path="table", \
                    **kwargs), masked=True)
            self.mask = astropy.table.Table.read(*args, path="mask", **kwargs)
        except OSError:
            self = Table(astropy.table.Table.read(*args, path="table",**kwargs))

        for col in self.colnames:
            if 'S' in self[col].dtype.str:
                self.replace_column(col, self[col].astype(str))

        return self

    def write_hdf5(self, *args, **kwargs):
        for col in self.colnames:
            if 'U' in self[col].dtype.str:
                self.replace_column(col, self[col].astype(bytes))

        self.write(*args, path='table', overwrite=True, **kwargs)

        for col in self.colnames:
            if 'S' in self[col].dtype.str:
                self.replace_column(col, self[col].astype(str))

        if self.masked:
            self.mask.write(*args, path='mask', append=True, **kwargs)

        return



import pdspy.modeling as modeling
import dynesty.plotting as dyplot
import matplotlib.pyplot as plt
import scipy.stats
import scipy.integrate
import pickle
import numpy

# Define a likelihood function.

def lnlike(p, visibilities, images, spectra, parameters, plot, \
        model="flared", ncpus=1, ncpus_highmass=1, with_hyperion=False, \
        timelimit=3600, source="ObjName", nice=19, verbose=False, \
        ftcode="galario"):

    # Set up the params dictionary.

    keys = []
    for key in sorted(parameters.keys()):
        if not parameters[key]["fixed"]:
            keys.append(key)

    params = dict(zip(keys, p))

    # Run the model.

    if model == "disk":
        m = modeling.run_disk_model(visibilities, images, spectra, params, \
                parameters, plot, ncpus=ncpus, ncpus_highmass=ncpus_highmass, \
                with_hyperion=with_hyperion, timelimit=timelimit, \
                source=source, nice=nice, verbose=verbose, ftcode=ftcode)
    else:
        m = modeling.run_flared_model(visibilities, params, parameters, plot, \
                ncpus=ncpus, source=source, nice=nice, ftcode=ftcode)

    # Catch whether the model timed out.

    if m == 0.:
        return -numpy.inf

    # A list to put all of the chisq into.

    chisq = []

    # Calculate the chisq for the visibilities.

    for j in range(len(visibilities["file"])):
        good = visibilities["data"][j].weights > 0

        chisq.append(-0.5*numpy.sum((visibilities["data"][j].real - \
                m.visibilities[visibilities["lam"][j]].real)**2 * \
                visibilities["data"][j].weights) - \
                numpy.sum(numpy.log(visibilities["data"][j].weights[good]/ \
                (2*numpy.pi))) + \
                -0.5*numpy.sum((visibilities["data"][j].imag - \
                m.visibilities[visibilities["lam"][j]].imag)**2 * \
                visibilities["data"][j].weights) - \
                numpy.sum(numpy.log(visibilities["data"][j].weights[good]/ \
                (2*numpy.pi))))

    # Calculate the chisq for all of the images.

    for j in range(len(images["file"])):
        chisq.append(-0.5 * (numpy.sum((images["data"][j].image - \
                m.images[images["lam"][j]].image)**2 / \
                images["data"][j].unc**2)))

    # Calculate the chisq for the SED.

    if "total" in spectra:
        chisq.append(-0.5 * (numpy.sum((spectra["total"].flux - \
                m.spectra["SED"].flux)**2 / spectra["total"].unc**2)))

    # Return the sum of the chisq.

    return numpy.array(chisq).sum()

# Define a prior function.

def ptform(u, parameters, priors, model="disk"):

    # Set up the params dictionary.

    keys = []
    for key in sorted(parameters.keys()):
        if not parameters[key]["fixed"]:
            keys.append(key)

    uparams = dict(zip(keys, u))
    pparams = dict(zip(keys, [0 for i in range(len(u))]))

    # Get the correct order for setting parameters (as some depend on others

    ordered_keys, index = numpy.unique(["logR_env","logR_disk","logR_in", \
            "logTatm0","logTmid0","logR_gap1"]+list(parameters.keys()), \
            return_index=True)
    ordered_keys = ordered_keys[numpy.argsort(index)]

    # Now loop through the parameters and transform the ones that aren't fixed.

    for key in ordered_keys:
        if not parameters[key]["fixed"]:
            # R_disk has to be smaller than R_env.
            if key == "logR_disk":
                if "logR_env" in pparams:
                    logR_env = pparams["logR_env"]
                else:
                    logR_env = parameters["logR_env"]["value"]

                pparams[key] = uparams[key] * (min(logR_env, \
                        parameters[key]["limits"][1]) - \
                        parameters[key]["limits"][0]) + \
                        parameters[key]["limits"][0]
            # R_in has to be smaller than R_disk.
            elif key == "logR_in":
                if "logR_disk" in pparams:
                    logR_disk = pparams["logR_disk"]
                else:
                    logR_disk = parameters["logR_disk"]["value"]

                pparams[key] = uparams[key] * (min(logR_disk, \
                        parameters[key]["limits"][1]) - \
                        parameters[key]["limits"][0]) + \
                        parameters[key]["limits"][0]
            # R_cav should be between R_in and R_disk.
            elif key == "logR_cav":
                if "logR_disk" in pparams:
                    logR_disk = pparams["logR_disk"]
                else:
                    logR_disk = parameters["logR_disk"]["value"]

                if "logR_in" in pparams:
                    logR_in = pparams["logR_in"]
                else:
                    logR_in = parameters["logR_in"]["value"]


                pparams[key] = uparams[key] * (min(pparams["logR_disk"], \
                        parameters[key]["limits"][1]) - \
                        max(pparams["logR_in"],parameters[key]["limits"][0])) +\
                        max(pparams["logR_in"],parameters[key]["limits"][0])
            # Same thing for R_gap and w_gap.
            elif key == "logR_gap1":
                if "logR_disk" in pparams:
                    logR_disk = pparams["logR_disk"]
                else:
                    logR_disk = parameters["logR_disk"]["value"]

                if "logR_in" in pparams:
                    logR_in = pparams["logR_in"]
                else:
                    logR_in = parameters["logR_in"]["value"]

                logR_gap_min = max(numpy.log10(10.**logR_in + \
                        parameters["w_gap1"]["limits"][0]/2), \
                        parameters["logR_gap1"]["limits"][0])
                logR_gap_max = min(numpy.log10(0.75*10.**logR_disk),
                        parameters["logR_gap1"]["limits"][1])

                pparams[key] = uparams[key] * (logR_gap_max - logR_gap_min) + \
                        logR_gap_min
            elif key == "w_gap1":
                if "logR_in" in pparams:
                    logR_in = pparams["logR_in"]
                else:
                    logR_in = parameters["logR_in"]["value"]

                if "logR_gap1" in pparams:
                    logR_gap = pparams["logR_gap1"]
                else:
                    logR_gap = parameters["logR_gap1"]["value"]

                w_gap_max = min(2*(10.**logR_gap - 10.**logR_in), \
                        parameters["w_gap1"]["limits"][1])

                pparams[key] = uparams[key] * (w_gap_max - \
                        parameters["w_gap1"]["limits"][0]) + \
                        parameters["w_gap1"]["limits"][0]
            # Disallow any absurdly dense models.
            elif key == "logR_env" and model == "disk":
                if params["logM_env"]["fixed"]:
                    logR_env_min = max(0.5*params["logM_env"]["value"] + 4., \
                            params["logR_env"]["limits"][0])
                else:
                    logR_env_min = params["logR_env"]["limits"][0]

                pparams[key] = uparams[key] * (parameters[key]["limits"][1] - \
                        logR_env_min) + logR_env_min
            elif key == "logM_env" and model == "disk":
                if "logR_env" in pparams:
                    logR_env = pparams["logR_env"]
                else:
                    logR_env = parameters["logR_env"]["value"]

                pparams[key] = uparams[key] * (min(2*(logR_env - 4), \
                        parameters[key]["limits"][1]) - \
                        parameters[key]["limits"][0]) + \
                        parameters[key]["limits"][0]
            # Tmid0 should be less than Tatm0.
            elif key == "logTmid0":
                pparams[key] = uparams[key] * (min(pparams["logTatm0"], \
                        parameters[key]["limits"][1]) - \
                        parameters[key]["limits"][0]) + \
                        parameters[key]["limits"][0]
            # Make the position angle cyclic.
            elif key == "pa":
                pparams[key] = (uparams[key] % 1.) * 360.
            # If we have a prior on the parallax, use that to get dpc.
            elif key == "dpc" and "parallax" in priors:
                m = priors["parallax"]["value"]
                s = priors["parallax"]["sigma"]
                low = 1. / parameters["dpc"]["limits"][1] * 1000
                high = 1. / parameters["dpc"]["limits"][0] * 1000
                low_n, high_n = (low - m) /s, (high - m) /s

                parallax = scipy.stats.truncnorm.ppf(uparams[key], low_n, \
                        high_n, loc=m, scale=s)

                pparams[key] = 1./parallax * 1000
            # If we have a prior on the stellar mass from the IMF.
            elif key == "logM_star" and "Mstar" in priors:
                imf = imf_gen(a=10.**parameters[key]["limits"][0], \
                        b=10.**parameters[key]["limits"][1])

                pparams[key] = numpy.log10(imf.ppf(uparams[key]))
            # If we have a prior on a parameter, draw the parameter from a 
            # normal distribution.
            elif key in priors:
                m = priors[key]["value"]
                s = priors[key]["sigma"]
                low = parameters[key]["limits"][0]
                high = parameters[key]["limits"][1]
                low_n, high_n = (low - m) /s, (high - m) /s

                pparams[key] = scipy.stats.truncnorm.ppf(uparams[key], low_n, \
                        high_n, loc=m, scale=s)
            # If none of the above apply, then draw from a uniform prior between
            # the provided limits.
            else:
                pparams[key] = uparams[key] * (parameters[key]["limits"][1] - \
                        parameters[key]["limits"][0]) + \
                        parameters[key]["limits"][0]

    # Now get the list of parameter values and return.

    p = [pparams[key] for key in sorted(keys)]

    return p

# Define a few useful classes for generating samples from the IMF.

def chabrier_imf(Mstar):
    if Mstar <= 1.:
        return 0.158 * 1./(numpy.log(10.) * Mstar) * \
                numpy.exp(-(numpy.log10(Mstar)-numpy.log10(0.08))**2/ \
                (2*0.69**2))
    else:
        return 4.43e-2 * Mstar**-1.3 * 1./(numpy.log(10.) * Mstar)

class imf_gen(scipy.stats.rv_continuous):
    def __init__(self, a=None, b=None):
        self.norm = scipy.integrate.quad(chabrier_imf, a, b)[0]

        super().__init__(a=a, b=b)

    def _pdf(self, x):
        return chabrier_imf(x) / self.norm

# Functions for saving the state of the Dynesty Sampler and loading a saved 
# state.

def save_sampler(name, sampler, pool=None, dynamic=False):

    # Clear the random state, as it cannot be pickled.
    sampler.rstate = None

    # Clear the MPI pool, as it also cannot be pickled.
    sampler.pool = None
    sampler.M = map

    # If this is a DynamicNestedSampler object, also do the sub-sampler.
    if dynamic:
        sampler.sampler.rstate = None
        sampler.sampler.pool = None
        sampler.sampler.M = map

    # Save
    pickle.dump(sampler, open(name, "wb"))

    # Restore everything to the way it was before.
    sampler.rstate = numpy.random
    sampler.pool = pool
    if pool != None:
        sampler.M = pool.map
    else:
        sampler.M = map

    # Again, repeat for the sub-sampler, if DynamicNestedSampler.

    if dynamic:
        sampler.sampler.rstate = numpy.random
        sampler.sampler.pool = pool
        if pool != None:
            sampler.sampler.M = pool.map
        else:
            sampler.sampler.M = map

def load_sampler(name, pool=None, dynamic=False):
    # Load the sampler from the pickle file.
    sampler = pickle.load(open("sampler.p","rb"))

    # Add back in the random state.
    sampler.rstate = numpy.random

    # Add the pool correctly.
    sampler.pool = pool
    if pool != None:
        sampler.M = pool.map
        sampler.queue_size = pool.size
    else:
        sampler.M = map

    # Add pool/random state correctly for the sub-sampler as well for 
    # the DynamicNestedSampler class.
    if dynamic:
        sampler.sampler.rstate = numpy.random
        sampler.sampler.pool = pool
        if pool != None:
            sampler.sampler.M = pool.map
            sampler.sampler.queue_size = pool.size
        else:
            sampler.sampler.M = map

    return sampler

# A function to make useful plots as the sampling is running.

def plot_status(res, ptform=None, labels=None, periodic=None):
    # Generate a plot of the trace.

    try:
        fig, ax = dyplot.traceplot(res, show_titles=True, trace_cmap="viridis",\
                connect=True, connect_highlight=range(5), labels=labels)
    except:
        # If it hasn't converged enough...
        fig, ax = dyplot.traceplot(res, show_titles=True, trace_cmap="viridis",\
                connect=True, connect_highlight=range(5), labels=labels, \
                kde=False)

    fig.savefig("traceplot.png")

    plt.close(fig)

    # Generate a bounds cornerplot.

    fig, ax = dyplot.cornerbound(res, it=res.niter-1, periodic=periodic, \
            prior_transform=ptform, show_live=True, labels=labels)

    fig.savefig("boundplot.png")

    plt.close(fig)



from .libinterferometry import Visibilities
import numpy

def rotate(data, pa=0):

    newu =  data.u * numpy.cos(pa) + data.v * numpy.sin(pa)
    newv = -data.u * numpy.sin(pa) + data.v * numpy.cos(pa)

    return Visibilities(newu, newv, data.freq.copy(), data.real.copy(), \
            data.imag.copy(), data.weights.copy())



from astropy.io.fits import open
from ..constants.physics import c
from .libinterferometry import Visibilities
import numpy

def readuvfits(filename, fmt="casa", fast=False):
    
    data = open(filename)
    
    header = data[0].header
    u = data[0].data.field(0).astype(numpy.float64)
    v = data[0].data.field(1).astype(numpy.float64)

    order = numpy.argsort(u)

    u = u[order]
    v = v[order]

    if fmt in ["casa","noema"]:
        arr = data[0].data.field("data").astype(numpy.float)

        for i in range(min(2,data[0].data.field("data").shape[5])):
            if i == 0:
                real = [arr[:,0,0,j,:,0,0] for j in range(arr.shape[3])]
                imag = [arr[:,0,0,j,:,0,1] for j in range(arr.shape[3])]
                weights = [arr[:,0,0,j,:,0,2] for j in range(arr.shape[3])]
                baselines = data[0].data.field(5)

                real = numpy.concatenate(real, axis=1)
                imag = numpy.concatenate(imag, axis=1)
                weights = numpy.concatenate(weights, axis=1)

                real = real[order,:]
                imag = imag[order,:]
                weights = weights[order,:]
                baselines = baselines[order]
            else:
                new_real = [arr[:,0,0,j,:,1,0] for j in range(arr.shape[3])]
                new_imag = [arr[:,0,0,j,:,1,1] for j in range(arr.shape[3])]
                new_weights = [arr[:,0,0,j,:,1,2] for j in range(arr.shape[3])]

                new_real = numpy.concatenate(new_real, axis=1)
                new_imag = numpy.concatenate(new_imag, axis=1)
                new_weights = numpy.concatenate(new_weights, axis=1)

                new_real = new_real[order,:]
                new_imag = new_imag[order,:]
                new_weights = new_weights[order,:]

                real = real*weights + new_real*new_weights
                imag = imag*weights + new_imag*new_weights
                weights += new_weights
                real[weights != 0] /= weights[weights != 0]
                imag[weights != 0] /= weights[weights != 0]
    elif fmt == "miriad":
        real = (data[0].data.field("data"))[:,0,0,:,0,0].astype(numpy.float64)
        imag = (data[0].data.field("data"))[:,0,0,:,0,1].astype(numpy.float64)
        weights = (data[0].data.field("data"))[:,0,0,:,0,2].\
                astype(numpy.float64)
        baselines = data[0].data.field(3)
    ant2 = numpy.mod(baselines,256)
    ant1 = (baselines-ant2)/256
    baseline = numpy.repeat("  6.1-6.1",u.size)
    baseline[((ant1 < 7) & (ant2 >= 7)) ^ ((ant1 >= 7) & (ant2 < 7))] = \
            " 6.1-10.4"
    baseline[(ant1 < 7) & (ant2 < 7)] = "10.4-10.4"
    
    u = numpy.concatenate((u, -u))
    v = numpy.concatenate((v, -v))
    real = numpy.concatenate((real, real))
    imag = numpy.concatenate((imag, -imag))
    weights = numpy.concatenate((weights, weights))
    baseline = numpy.concatenate((baseline, baseline))
    
    if fmt == "casa":
        IF = data[1].data.field('if freq')[0]
        delta_freq = data[1].data.field('ch width')[0]
    else:
        IF = 0.
        delta_freq = header["CDELT4"]
    freq0 = header["CRVAL4"]
    pix0 = header["CRPIX4"]
    nfreq = header["NAXIS4"]

    if isinstance(IF, float):
        IF = numpy.array([IF])
        delta_freq = numpy.array([delta_freq])

    freq = [freq0 + IF[i] + (numpy.arange(nfreq)-(pix0-1))*delta_freq[i] \
            for i in range(IF.size)]

    freq = numpy.concatenate(freq)
    nfreq = header["NAXIS4"]*IF.size
    
    u *= freq.mean()
    v *= freq.mean()

    weights *= nfreq
    
    data.close()

    uvdata = Visibilities(u, v, freq, real, -imag, weights, baseline=baseline)
    
    uvdata.set_header(header)
    
    return uvdata
